"""
Setup:
Follow setup instruction from the github documentation to create personal access token
https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/creating-a-personal-access-token
"""

from typing import Annotated, Literal, Union
from httpx import Client
from dateutil import parser
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.utils.collections import is_not_empty, is_empty
from admyral.secret.secret import register_secret


@register_secret(secret_type="GitHub")
class GitHubSecret(BaseModel):
    access_token: str


def get_github_enterprise_client(secret: GitHubSecret, enterprise: str) -> Client:
    return Client(
        base_url=f"https://api.github.com/enterprises/{enterprise}",
        headers={
            "Authorization": f"Bearer {secret.access_token}",
            "Accept": "application/vnd.github.v3+json",
        },
    )


def get_github_client(secret: GitHubSecret) -> Client:
    return Client(
        base_url="https://api.github.com",
        headers={
            "Authorization": f"Bearer {secret.access_token}",
            "Accept": "application/vnd.github.v3+json",
        },
    )


def _get_events_with_pagination(
    client: Client,
    url: str,
    params: dict,
    limit: int | None = None,
    event_filter_fn: Union[callable, None] = None,
    early_stop_fn: Union[callable, None] = None,
) -> list[dict[str, JsonValue]]:
    events = []
    while limit is None or len(events) < limit:
        response = client.get(url, params=params)
        response.raise_for_status()
        events_on_page = response.json()

        if event_filter_fn:
            events.extend(filter(event_filter_fn, events_on_page))
        else:
            events.extend(events_on_page)

        if early_stop_fn and early_stop_fn(events_on_page):
            break

        if "next" in response.links:
            url = response.links["next"]["url"][len(str(client.base_url)) :]
            params = None
        else:
            break

    return events if limit is None else events[:limit]


@action(
    display_name="Search Enterprise Audit Logs",
    display_namespace="GitHub",
    description="Search GitHub audit logs for an enterprise.",
    secrets_placeholders=["GITHUB_ENTERPRISE_SECRET"],
)
def search_github_enterprise_audit_logs(
    enterprise: Annotated[
        str,
        ArgumentMetadata(
            display_name="Enterprise",
            description="The slug version of the enterprise name. You can also substitute this value with the enterprise id.",
        ),
    ],
    filter: Annotated[
        str,
        ArgumentMetadata(
            display_name="Filter",
            description="Filter to apply to the query.",
        ),
    ] = None,
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="Maximum number of logs to return.",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    # https://docs.github.com/en/enterprise-cloud@latest/rest/enterprise-admin/audit-log?apiVersion=2022-11-28#get-the-audit-log-for-an-enterprise
    # https://docs.github.com/en/search-github/getting-started-with-searching-on-github/understanding-the-search-syntax#query-for-dates

    secret = ctx.get().secrets.get("GITHUB_ENTERPRISE_SECRET")
    secret = GitHubSecret.model_validate(secret)

    with get_github_enterprise_client(secret, enterprise) as client:
        params = {"order": "asc", "per_page": 100}

        phrases = []

        if filter:
            phrases.append(filter)

        if start_time and end_time:
            phrases.append(f"created:{start_time}..{end_time}")
        elif start_time:
            phrases.append(f"created:>={start_time}")
        elif end_time:
            phrases.append(f"created:<={end_time}")

        if is_not_empty(phrases):
            params["phrase"] = " ".join(phrases)

        events = _get_events_with_pagination(
            client=client,
            url="/audit-log",
            params=params,
            limit=limit,
        )

        return events if limit is None else events[:limit]


@action(
    display_name="Compare Two GitHub Commits",
    display_namespace="GitHub",
    description="Get the commit diff info between two commits.",
    secrets_placeholders=["GITHUB_SECRET"],
)
def compare_two_github_commits(
    repo_owner: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Owner",
            description="The owner of the repository",
        ),
    ],
    repo_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Name",
            description="The name of the repository",
        ),
    ],
    base: Annotated[
        str,
        ArgumentMetadata(
            display_name="Base",
            description="The base commit (commit ID or SHA)",
        ),
    ],
    head: Annotated[
        str,
        ArgumentMetadata(
            display_name="Head",
            description="The head commit (commit ID or SHA)",
        ),
    ],
    diff_type: Annotated[
        Literal["json", "diff"],
        ArgumentMetadata(
            display_name="Diff Type",
            description="The type of diff to return. 'json' returns the diff as JSON. "
            "'diff' returns the diff as string in the same format as the CLI command "
            "git diff BASE..HEAD. Default: 'json'.",
        ),
    ] = "json",
) -> dict[str, JsonValue]:
    # https://docs.github.com/en/rest/commits/commits?apiVersion=2022-11-28#compare-two-commits

    secret = ctx.get().secrets.get("GITHUB_SECRET")
    secret = GitHubSecret.model_validate(secret)

    with get_github_client(secret) as client:
        url = f"/repos/{repo_owner}/{repo_name}/compare/{base}...{head}"

        if diff_type == "diff":
            headers = {"Accept": "application/vnd.github.v3.diff"}
            diff = ""
            while True:
                response = client.get(url, headers=headers)
                response.raise_for_status()
                diff += response.text

                if "next" in response.links:
                    url = response.links["next"]["url"][len(str(client.base_url)) :]
                else:
                    break

            return diff

        response = client.get(url)
        response.raise_for_status()

        return response.json()


@action(
    display_name="List Merged Pull Requests for a Repository",
    display_namespace="GitHub",
    description="List all merged pull requests for a repository.",
    secrets_placeholders=["GITHUB_SECRET"],
)
def list_github_merged_pull_requests(
    repo_owner: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Owner",
            description="The owner of the repository",
        ),
    ],
    repo_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Name",
            description="The name of the repository",
        ),
    ],
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    # https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28

    secret = ctx.get().secrets.get("GITHUB_SECRET")
    secret = GitHubSecret.model_validate(secret)

    with get_github_client(secret) as client:
        params = {
            "state": "closed",
            "sort": "updated",
            "direction": "desc",
            "per_page": 100,
        }

        # convert into datetime object
        start_time = parser.isoparse(
            start_time if start_time else "1970-01-01T00:00:00Z"
        )
        end_time = parser.isoparse(end_time if end_time else "2100-01-01T00:00:00Z")

        events = _get_events_with_pagination(
            client=client,
            url=f"/repos/{repo_owner}/{repo_name}/pulls",
            params=params,
            limit=limit,
            event_filter_fn=lambda event: event["merged_at"]
            and start_time <= parser.isoparse(event["merged_at"]) <= end_time,
            early_stop_fn=lambda events: end_time
            < parser.isoparse(events[-1]["updated_at"]),
        )

        return events


@action(
    display_name="Get Commit History for a Pull Request",
    display_namespace="GitHub",
    description="List commit history for a Pull Request from most recent to oldest",
    secrets_placeholders=["GITHUB_SECRET"],
)
def list_github_commit_history_for_pull_request(
    repo_owner: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Owner",
            description="The owner of the repository",
        ),
    ],
    repo_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Name",
            description="The name of the repository",
        ),
    ],
    pull_request_number: Annotated[
        int,
        ArgumentMetadata(
            display_name="Pull Request Number",
            description="The name of the repository",
        ),
    ],
) -> list[dict[str, JsonValue]]:
    # https://docs.github.com/en/rest/commits/commits?apiVersion=2022-11-28

    secret = ctx.get().secrets.get("GITHUB_SECRET")
    secret = GitHubSecret.model_validate(secret)

    with get_github_client(secret) as client:
        events = _get_events_with_pagination(
            client=client,
            url=f"/repos/{repo_owner}/{repo_name}/pulls/{pull_request_number}/commits",
            params={"per_page": 100},
        )
        return events


@action(
    display_name="List Review History for a Pull Request",
    display_namespace="GitHub",
    description="List review history for a pull request.",
    secrets_placeholders=["GITHUB_SECRET"],
)
def list_github_review_history_for_pull_request(
    repo_owner: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Owner",
            description="The owner of the repository",
        ),
    ],
    repo_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Name",
            description="The name of the repository",
        ),
    ],
    pull_request_number: Annotated[
        int,
        ArgumentMetadata(
            display_name="Pull Request Number",
            description="The name of the repository",
        ),
    ],
    state: Annotated[
        Literal["APPROVED", "CHANGES_REQUESTED", "COMMENTED", "DISMISSED"] | None,
        ArgumentMetadata(
            display_name="State",
            description="The state of the review to list.",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    # https://docs.github.com/en/rest/pulls/reviews?apiVersion=2022-11-28
    secret = ctx.get().secrets.get("GITHUB_SECRET")
    secret = GitHubSecret.model_validate(secret)

    with get_github_client(secret) as client:
        params = {
            "per_page": 100,
        }

        if state:
            event_filter_fn = lambda event: event["state"] == state  # noqa: E731
        else:
            event_filter_fn = None

        events = _get_events_with_pagination(
            client=client,
            url=f"/repos/{repo_owner}/{repo_name}/pulls/{pull_request_number}/reviews",
            params=params,
            event_filter_fn=event_filter_fn,
        )

        return events


@action(
    display_name="List Merged Pull Requests Without Approval",
    display_namespace="GitHub",
    description="List all pull requests of a repository that contain unapproved commits, i.e., PRs which were never approved or commits after an approval.",
    secrets_placeholders=["GITHUB_SECRET"],
)
def list_github_merged_pull_requests_without_approval(
    repo_owner: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Owner",
            description="The owner of the repository",
        ),
    ],
    repo_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Name",
            description="The name of the repository",
        ),
    ],
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    merged_prs = list_github_merged_pull_requests(
        repo_owner=repo_owner,
        repo_name=repo_name,
        start_time=start_time,
        end_time=end_time,
    )

    unreviewed_prs = []
    for pr in merged_prs:
        commit_history = list_github_commit_history_for_pull_request(
            repo_owner=repo_owner, repo_name=repo_name, pull_request_number=pr["number"]
        )
        if is_empty(commit_history):
            continue

        # Identify the latest commit
        last_commit = sorted(
            commit_history,
            key=lambda commit: parser.parse(commit["commit"]["committer"]["date"]),
            reverse=True,
        )[0]
        last_commit_id = last_commit["sha"]

        # Identify the last approved commit
        approval_history = list_github_review_history_for_pull_request(
            repo_owner=repo_owner,
            repo_name=repo_name,
            pull_request_number=pr["number"],
            state="APPROVED",
        )
        if is_empty(approval_history):
            unreviewed_prs.append(
                {
                    "number": pr["number"],
                    "title": pr["title"],
                    "html_url": pr["html_url"],
                    "user": pr["user"]["login"],
                    "merged_at": pr["merged_at"],
                    "last_commit_id": last_commit_id,
                    "last_approved_commit_id": pr["base"]["sha"],
                }
            )
            continue

        approval_history = sorted(
            approval_history,
            key=lambda approval: parser.parse(approval["submitted_at"]),
            reverse=True,
        )

        last_approved_commit_id = approval_history[0]["commit_id"]
        if last_commit_id != last_approved_commit_id:
            unreviewed_prs.append(
                {
                    "number": pr["number"],
                    "title": pr["title"],
                    "html_url": pr["html_url"],
                    "user": pr["user"]["login"],
                    "merged_at": pr["merged_at"],
                    "last_commit_id": last_commit_id,
                    "last_approved_commit_id": last_approved_commit_id,
                }
            )

    return unreviewed_prs


@action(
    display_name="Get Pull Request Comments",
    display_namespace="GitHub",
    description="Get comments for a pull request.",
    secrets_placeholders=["GITHUB_SECRET"],
)
def list_github_issue_comments(
    repo_owner: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Owner",
            description="The owner of the repository",
        ),
    ],
    repo_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Repository Name",
            description="The name of the repository",
        ),
    ],
    number: Annotated[
        int,
        ArgumentMetadata(
            display_name="Pull Request/Issue Number",
            description="The number of the pull request or issue",
        ),
    ],
) -> list[dict[str, JsonValue]]:
    # https://docs.github.com/en/rest/issues/comments?apiVersion=2022-11-28#list-issue-comments
    secret = ctx.get().secrets.get("GITHUB_SECRET")
    secret = GitHubSecret.model_validate(secret)

    with get_github_client(secret) as client:
        params = {
            "per_page": 100,
        }

        events = _get_events_with_pagination(
            client=client,
            url=f"/repos/{repo_owner}/{repo_name}/issues/{number}/comments",
            params=params,
        )

        return events



from typing import Optional
from fastapi import APIRouter, status, Depends

from admyral.server.auth import authenticate
from admyral.models import PythonAction, AuthenticatedUser, ActionMetadata
from admyral.server.deps import get_admyral_store


router = APIRouter()


@router.post("/push", status_code=status.HTTP_204_NO_CONTENT)
async def push_action(
    python_action: PythonAction,
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> None:
    """
    Push a Python action to the store. If the action for the provided action type already exists for the user,
    it will be overwritten.

    Args:
        python_action: The Python action object.
    """
    await get_admyral_store().store_action(
        user_id=authenticated_user.user_id, action=python_action
    )


@router.get("", status_code=status.HTTP_200_OK)
async def list_actions(
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> list[ActionMetadata]:
    """
    List all stored custom Python actions.

    Returns:
        A list of stored custom Python action objects.
    """
    return await get_admyral_store().list_actions(user_id=authenticated_user.user_id)


@router.delete("/{action_type}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_action(
    action_type: str, authenticated_user: AuthenticatedUser = Depends(authenticate)
) -> None:
    """
    Delete a stored custom Python action by its action type.

    Args:
        action_type: function name of the custom Python action.
    """
    await get_admyral_store().delete_action(
        action_type=action_type, user_id=authenticated_user.user_id
    )


@router.get("/{action_type}", status_code=status.HTTP_200_OK)
async def get_action(
    action_type: str, authenticated_user: AuthenticatedUser = Depends(AuthenticatedUser)
) -> Optional[PythonAction]:
    """
    Get a Python action by its action type.

    Args:
        action_type: The action type.

    Returns:
        The Python action object.
    """
    return await get_admyral_store().get_action(
        user_id=authenticated_user.user_id, action_type=action_type
    )



from admyral.actions.ai_action import ai_action
from admyral.actions.send_email import send_email
from admyral.actions.utilities import (
    deserialize_json_string,
    serialize_json_string,
    wait,
    transform,
    get_time_interval_of_last_n_days,
    get_time_interval_of_last_n_hours,
    format_json_to_list_view_string,
    send_to_workflow,
    send_list_elements_to_workflow,
    split_text,
    build_lookup_table,
    filter,
    join_lists,
    select_fields_from_objects_in_list,
)
from admyral.actions.integrations.communication import (
    send_slack_message,
    lookup_slack_user_by_email,
    send_slack_message_to_user_by_email,
    batched_send_slack_message_to_user_by_email,
)
from admyral.actions.integrations.enrich import (
    alienvault_otx_analyze_domain,
    grey_noise_ip_lookup,
    virus_total_analyze_hash,
    virus_total_analyze_domain,
    virus_total_analyze_ip,
    virus_total_analyze_url,
    abuseipdb_analyze_ip,
    leakcheck_v2_lookup,
    leakcheck_public_lookup,
)
from admyral.actions.integrations.cases import (
    create_jira_issue,
    update_jira_issue_status,
    comment_jira_issue_status,
    search_jira_issues,
    get_jira_audit_records,
    create_opsgenie_alert,
    create_pagerduty_incident,
    get_jira_project,
    get_jira_transitions,
)
from admyral.actions.integrations.vulnerability_management import (
    list_snyk_vulnerabilities,
    list_amazon_inspector2_vulnerabilities,
)
from admyral.actions.integrations.ai import (
    anthropic_chat_completion,
    azure_openai_chat_completion,
    openai_chat_completion,
    mistralai_chat_completion,
)
from admyral.actions.integrations.edr import (
    list_ms_defender_for_endpoint_alerts,
    list_sentinel_one_alerts,
)
from admyral.actions.integrations.cdr import (
    list_ms_defender_for_cloud_alerts,
    list_wiz_alerts,
)
from admyral.actions.integrations.email import list_abnormal_security_cases
from admyral.actions.integrations.siem import list_ms_sentinel_alerts
from admyral.actions.integrations.iam import (
    list_okta_events,
    okta_search_users,
    okta_get_all_user_types,
    get_okta_logs,
)
from admyral.actions.integrations.compliance import (
    list_retool_inactive_users,
    list_groups_per_user,
    list_1password_audit_events,
    search_github_enterprise_audit_logs,
    list_github_merged_pull_requests,
    list_github_commit_history_for_pull_request,
    list_github_review_history_for_pull_request,
    compare_two_github_commits,
    list_github_merged_pull_requests_without_approval,
    list_google_docs_revisions,
    list_ms_intune_unencrypted_managed_devices,
    list_ms_intune_managed_devices,
    list_github_issue_comments,
    list_kandji_devices,
    get_kandji_device_details,
    list_kandji_unencrypted_devices,
    get_kandji_device_apps,
    list_zendesk_users,
    get_kandji_application_firewall,
    get_kandji_desktop_and_screensaver,
    get_kandji_library_item_statuses,
    list_groups_and_apps_per_user,
    list_used_groups_and_apps_per_user,
    list_google_drive_files_with_link_sharing_enabled,
)
from admyral.actions.integrations.cloud import (
    steampipe_query_aws,
    aws_s3_bucket_logging_enabled,
    aws_s3_bucket_enforce_ssl,
    aws_s3_restrict_public_write_access,
    aws_s3_restrict_public_read_access,
    aws_s3_default_encryption_enabled,
    aws_list_iam_users,
)
from admyral.actions.integrations.database import run_sql_query


__all__ = [
    "ai_action",
    "send_email",
    "deserialize_json_string",
    "serialize_json_string",
    "wait",
    "transform",
    "send_slack_message",
    "lookup_slack_user_by_email",
    "send_slack_message_to_user_by_email",
    "batched_send_slack_message_to_user_by_email",
    "alienvault_otx_analyze_domain",
    "grey_noise_ip_lookup",
    "virus_total_analyze_hash",
    "virus_total_analyze_domain",
    "virus_total_analyze_ip",
    "virus_total_analyze_url",
    "create_jira_issue",
    "update_jira_issue_status",
    "comment_jira_issue_status",
    "search_jira_issues",
    "get_jira_audit_records",
    "create_opsgenie_alert",
    "create_pagerduty_incident",
    "list_snyk_vulnerabilities",
    "list_amazon_inspector2_vulnerabilities",
    "anthropic_chat_completion",
    "azure_openai_chat_completion",
    "openai_chat_completion",
    "mistralai_chat_completion",
    "list_abnormal_security_cases",
    "list_ms_defender_for_endpoint_alerts",
    "list_ms_defender_for_cloud_alerts",
    "list_ms_sentinel_alerts",
    "list_okta_events",
    "list_sentinel_one_alerts",
    "list_wiz_alerts",
    "list_retool_inactive_users",
    "list_groups_per_user",
    "list_1password_audit_events",
    "okta_search_users",
    "okta_get_all_user_types",
    "get_okta_logs",
    "search_github_enterprise_audit_logs",
    "list_github_merged_pull_requests",
    "list_github_commit_history_for_pull_request",
    "list_github_review_history_for_pull_request",
    "list_github_merged_pull_requests_without_approval",
    "compare_two_github_commits",
    "abuseipdb_analyze_ip",
    "steampipe_query_aws",
    "aws_s3_bucket_logging_enabled",
    "aws_s3_bucket_enforce_ssl",
    "aws_s3_restrict_public_write_access",
    "aws_s3_restrict_public_read_access",
    "aws_s3_default_encryption_enabled",
    "get_time_interval_of_last_n_days",
    "get_time_interval_of_last_n_hours",
    "format_json_to_list_view_string",
    "run_sql_query",
    "list_google_docs_revisions",
    "list_ms_intune_unencrypted_managed_devices",
    "list_ms_intune_managed_devices",
    "send_to_workflow",
    "send_list_elements_to_workflow",
    "split_text",
    "list_github_issue_comments",
    "list_kandji_devices",
    "get_kandji_device_details",
    "list_kandji_unencrypted_devices",
    "get_kandji_device_apps",
    "aws_list_iam_users",
    "list_zendesk_users",
    "build_lookup_table",
    "filter",
    "get_kandji_application_firewall",
    "get_kandji_desktop_and_screensaver",
    "get_kandji_library_item_statuses",
    "join_lists",
    "leakcheck_v2_lookup",
    "leakcheck_public_lookup",
    "list_groups_and_apps_per_user",
    "list_used_groups_and_apps_per_user",
    "list_google_drive_files_with_link_sharing_enabled",
    "get_jira_project",
    "get_jira_transitions",
    "select_fields_from_objects_in_list",
]



from typing import Annotated, Literal
from httpx import Client, HTTPStatusError
from collections import defaultdict
from dateutil import parser
from datetime import datetime, timezone, timedelta
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_exception_type,
)
import itertools
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.exceptions import NonRetryableActionError, RetryableActionError
from admyral.secret.secret import register_secret


@register_secret(secret_type="Retool")
class RetoolSecret(BaseModel):
    domain: str
    api_key: str


def get_retool_client(secret: RetoolSecret) -> Client:
    # Auth: https://docs.retool.com/org-users/guides/retool-api/authentication#tag/Organization/paths/~1usage~1organizations/get
    return Client(
        base_url=f"https://{secret.domain}/api/v2",
        headers={
            "Authorization": f"Bearer {secret.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@retry(
    stop=stop_after_attempt(15),
    wait=wait_random_exponential(multiplier=1, max=90),
    retry=retry_if_exception_type(RetryableActionError),  # Don't retry on 4xx errors
    reraise=True,
)
def _list_retool_api(
    client: Client,
    url: str,
    next_token: str | None = None,
    method: Literal["get", "post"] = "get",
    json_body: dict = {},
    params: dict = {},
) -> dict[str, JsonValue]:
    if next_token is not None:
        params["next_token"] = next_token

    try:
        if method == "post":
            response = client.post(url, json=json_body, params=params)
        else:
            response = client.get(url, params=params)
        response.raise_for_status()
        response_json = response.json()

        if not response_json["success"]:
            raise NonRetryableActionError(
                f"Failed to call Retool API: {response_json['message']}"
            )

        return response_json

    except HTTPStatusError as e:
        if e.response.status_code == 429:  # Rate limit exceeded
            raise RetryableActionError("Rate limit exceeded")
        if e.response.status_code >= 500:  # Server error
            raise RetryableActionError(f"Server error: {e.response.text}")
        raise NonRetryableActionError(f"Failed to call Retool API: {e.response.text}")


def _list_retool_api_with_pagination(
    client: Client,
    url: str,
    method: Literal["get", "post"] = "get",
    json_body: dict = {},
) -> list[dict[str, JsonValue]]:
    data = []

    next_token = None
    while True:
        response = _list_retool_api(client, url, next_token, method, json_body)
        data.extend(response["data"])

        next_token = response.get("next_token")
        if not response.get("has_more"):
            break

    return data


@action(
    display_name="List Inactive Retool Users",
    display_namespace="Retool",
    description="This method lists Retool users who have not logged in for a specified number of days.",
    secrets_placeholders=["RETOOL_SECRET"],
)
def list_retool_inactive_users(
    inactivity_threshold_in_days: Annotated[
        int,
        ArgumentMetadata(
            display_name="Inactivity Threshold in Days",
            description="The number of days since the last login after which a user is considered inactive.",
        ),
    ] = 60,
) -> list[dict[str, JsonValue]]:
    # Returns a list of users. The API token must have the "Users > Read" scope.
    # https://docs.retool.com/reference/api/#tag/Users/paths/~1users~1%7BuserId%7D/patch

    secret = ctx.get().secrets.get("RETOOL_SECRET")
    secret = RetoolSecret.model_validate(secret)

    with get_retool_client(secret) as client:
        users = _list_retool_api_with_pagination(client, "/users")
        utc_time_now = datetime.now(timezone.utc) - timedelta(
            days=inactivity_threshold_in_days
        )
        return [
            user
            for user in users
            if parser.isoparse(user["last_active"]) <= utc_time_now
        ]


@action(
    display_name="List Groups per User",
    display_namespace="Retool",
    description="List all groups a user is a member of as well as the last time a user was active in Retool.",
    secrets_placeholders=["RETOOL_SECRET"],
)
def list_groups_per_user() -> dict[str, JsonValue]:
    # Get all permission groups for an organization or space. The API token must have the "Groups > Read" scope.
    # https://docs.retool.com/reference/api/#tag/Users/paths/~1users~1%7BuserId%7D/patch
    # https://docs.retool.com/reference/api/v2/#tag/Groups/paths/~1groups~1%7BgroupId%7D/put

    secret = ctx.get().secrets.get("RETOOL_SECRET")
    secret = RetoolSecret.model_validate(secret)

    with get_retool_client(secret) as client:
        # Fetch all users to get the last active date
        users = _list_retool_api_with_pagination(client, "/users")

        groups_per_user = {
            user["email"]: {
                "groups": [],
                "last_active": user["last_active"],
            }
            for user in users
        }

        # Fetch all groups
        groups = _list_retool_api_with_pagination(client, "/groups")
        for group in groups:
            for member in group["member"]:
                groups_per_user[member]["groups"].append(group["name"])

        return groups_per_user


@action(
    display_name="List Groups and Apps per User",
    display_namespace="Retool",
    description="List all groups a user is a member of and all the apps a user has access to.",
    secrets_placeholders=["RETOOL_SECRET"],
)
def list_groups_and_apps_per_user() -> list[dict[str, JsonValue]]:
    secret = ctx.get().secrets.get("RETOOL_SECRET")
    secret = RetoolSecret.model_validate(secret)

    with get_retool_client(secret) as client:
        # https://docs.retool.com/reference/api/#tag/Users/paths/~1users/get
        users = _list_retool_api_with_pagination(client, "/users")

        user_to_groups_and_apps = {
            user["email"]: {
                "user": user["email"],
                "groups": [],
                "apps_own_access": set(),
                "apps_edit_access": set(),
                "apps_use_access": set(),
            }
            for user in users
        }

        # https://docs.retool.com/reference/api/#tag/Groups/paths/~1groups/get
        groups = _list_retool_api_with_pagination(client, "/groups")

        # use cache to avoid hitting the /apps/{appId} endpoint multiple times
        app_details_cache = {}

        for group in groups:
            # List apps (object_type == "app") a group can access
            # https://docs.retool.com/reference/api/#tag/Groups/paths/~1permissions~1listObjects/post
            body = {
                "subject": {
                    "type": "group",
                    "id": group["id"],
                },
                "object_type": "app",
            }
            group_accessible_apps = _list_retool_api_with_pagination(
                client, "/permissions/listObjects", method="post", json_body=body
            )

            # fetch the app name
            apps_by_access_level = defaultdict(list)
            VALID_ACCESS_LEVELS = {"own", "edit", "use"}

            for app_access in group_accessible_apps:
                app_id = app_access["id"]
                access_level = app_access["access_level"]
                if access_level not in VALID_ACCESS_LEVELS:
                    raise ValueError(
                        f"Invalid access level: {access_level}. Expected one of: {VALID_ACCESS_LEVELS}"
                    )

                # https://docs.retool.com/reference/api/#tag/Apps/paths/~1apps~1%7BappId%7D/get
                app_details = app_details_cache.get(app_id)
                if app_details is None:
                    app_details = _list_retool_api(client, f"/apps/{app_id}")
                    app_details_cache[app_id] = app_details

                apps_by_access_level[access_level].append(app_details["name"])

            # merge with user information
            for member in group["members"]:
                user_to_groups_and_apps[member]["groups"].append(group["name"])
                user_to_groups_and_apps[member]["apps_own_access"].update(
                    apps_by_access_level.get("own", [])
                )
                user_to_groups_and_apps[member]["apps_edit_access"].update(
                    apps_by_access_level.get("edit", [])
                )
                user_to_groups_and_apps[member]["apps_use_access"].update(
                    apps_by_access_level.get("use", [])
                )

        return list(
            {
                "user": user["user"],
                "groups": user["groups"],
                "apps_own_access": list(user["apps_own_access"]),
                "apps_edit_access": list(user["apps_edit_access"]),
                "apps_use_access": list(user["apps_use_access"]),
            }
            for user in user_to_groups_and_apps.values()
        )


def _validate_date_format(date_str: str) -> str:
    try:
        datetime.strptime(date_str, "%Y-%m-%d")
        return date_str
    except ValueError:
        raise ValueError(
            f"Invalid date format: {date_str}. Expected format: YYYY-MM-DD"
        )


@action(
    display_name="List Used Groups and Apps per User",
    display_namespace="Retool",
    description="List all apps a user has used as viewer and/or editor within a certain time range.",
    secrets_placeholders=["RETOOL_SECRET"],
)
def list_used_groups_and_apps_per_user(
    start_date: Annotated[
        str,
        ArgumentMetadata(
            display_name="Start Date",
            description="The start date of the time range. Example: 2022-01-01",
        ),
    ],
    end_date: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Date",
            description="The end date of the time range. If not specified, then minimum(start_date + 30 days, today - 1) is used. Example: 2022-12-31",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    secret = ctx.get().secrets.get("RETOOL_SECRET")
    secret = RetoolSecret.model_validate(secret)

    start_date = _validate_date_format(start_date)
    if end_date is not None:
        end_date = _validate_date_format(end_date)

    with get_retool_client(secret) as client:
        # https://docs.retool.com/reference/api/#tag/Users/paths/~1users/get
        users = _list_retool_api_with_pagination(client, "/users")

        app_id_to_groups = defaultdict(list)
        # https://docs.retool.com/reference/api/#tag/Groups/paths/~1groups/get
        groups = _list_retool_api_with_pagination(client, "/groups")
        for group in groups:
            # List apps (object_type == "app") a group can access
            # https://docs.retool.com/reference/api/#tag/Groups/paths/~1permissions~1listObjects/post
            body = {
                "subject": {
                    "type": "group",
                    "id": group["id"],
                },
                "object_type": "app",
            }
            apps = _list_retool_api_with_pagination(
                client, "/permissions/listObjects", method="post", json_body=body
            )

            for app in apps:
                app_id_to_groups[app["id"]].append(group["name"])

        result = []
        for user in users:
            # fetch the app usage for the user
            # https://docs.retool.com/reference/api/#tag/Usage/paths/~1usage~1user_details/get
            params = {
                "start_date": start_date,
                "email": user["email"],
            }
            if end_date is not None:
                params["end_date"] = end_date
            response = _list_retool_api(client, "/usage/user_details", params=params)
            user_details = response["data"]

            # calculate the groups the user has used
            used_groups = set(
                itertools.chain.from_iterable(
                    app_id_to_groups[app["id"]]
                    for app in user_details["editor_summary"]
                )
            ) | set(
                itertools.chain.from_iterable(
                    app_id_to_groups[app["id"]]
                    for app in user_details["viewer_summary"]
                )
            )
            used_groups = list(used_groups)

            result.append(
                {
                    "user": user["email"],
                    "viewed_apps": list(
                        app["app_name"] for app in user_details["viewer_summary"]
                    ),
                    "edited_apps": list(
                        app["app_name"] for app in user_details["editor_summary"]
                    ),
                    "used_groups": used_groups,
                }
            )

        return result



import ast

from admyral.models import (
    Condition,
)


def compile_condition_str(condition: str) -> Condition:
    """
    Compile condition string to condition expression AST.

    Note: The string should not contain any references but be a pure pythonic condition expression.

    Args:
        condition: Condition string.

    Returns:
        Condition expression AST.
    """
    from admyral.compiler.workflow_compiler import WorkflowCompiler

    return WorkflowCompiler().compile_if_condition_str(condition)


def condition_to_str(condition: ast.expr) -> str:
    """
    Convert condition expression AST to string.

    Args:
        condition: Condition expression AST.

    Returns:
        Condition string.
    """
    return ast.unparse(condition)



import requests
from admyral.typings import JsonValue

from admyral.models import (
    Workflow,
    WorkflowDAG,
    PythonAction,
    WorkflowPushResponse,
    Secret,
    WorkflowPushRequest,
    WorkflowTriggerResponse,
    SecretMetadata,
    ActionMetadata,
)
from admyral.config.config import API_V1_STR


class AdmyralClient:
    def __init__(
        self, base_url: str = "http://localhost:8000", api_key: str | None = None
    ) -> None:
        self.base_url = base_url
        self.api_key = api_key

    def _request(
        self,
        method: str,
        path: str,
        params: dict | None = None,
        json: JsonValue | None = None,
        webhook_secret: str | None = None,
    ) -> None | JsonValue:
        headers = {
            "Content-Type": "application/json",
        }
        if webhook_secret:
            headers["Authorization"] = webhook_secret
        elif self.api_key:
            headers["x-api-key"] = self.api_key

        response = None
        if method == "GET":
            response = requests.get(
                f"{self.base_url}{path}",
                headers=headers,
                params=params,
            )
        elif method == "POST":
            response = requests.post(
                f"{self.base_url}{path}", headers=headers, json=json, params=params
            )
        elif method == "DELETE":
            response = requests.delete(
                f"{self.base_url}{path}", headers=headers, json=json, params=params
            )
        else:
            raise NotImplementedError(f"Missing implementation for method: {method}")

        if response.status_code == 401:
            raise RuntimeError("Unauthorized. Please check your API key.")

        if not response.ok:
            error_message = response.text
            raise RuntimeError(
                f"Request failed with status code {response.status_code}. Error: {error_message}"
            )

        return response.json() if response.text else None

    def _get(self, path: str, params: dict = {}) -> None | JsonValue:
        return self._request("GET", path, params)

    def _post(
        self, path: str, json: JsonValue | None = None, params: dict = {}
    ) -> None | JsonValue:
        return self._request("POST", path, json=json, params=params)

    def _delete(
        self, path: str, json: JsonValue | None = None, params: dict = {}
    ) -> None | JsonValue:
        return self._request("DELETE", path, json=json, params=params)

    ########################################################
    # Workflows
    ########################################################

    def get_workflow(self, workflow_name: str) -> Workflow | None:
        """
        Returns the workflow with the given id if it exists.

        Args:
            workflow_name: The workflow name.

        Returns:
            The workflow with the given name if it exists, otherwise None.
        """
        result = self._get(f"{API_V1_STR}/workflows/{workflow_name}")
        return Workflow.model_validate(result) if result else None

    def push_workflow(
        self, workflow_name: str, workflow_dag: WorkflowDAG, is_active: bool = False
    ) -> WorkflowPushResponse:
        """
        Pushes the workflow to the server.

        Args:
            workflow_name: The workflow to push.

        Returns:
            The response from the server which contains the webhook id and secret
            if the workflow has webhook enabled.
        """
        response = self._post(
            f"{API_V1_STR}/workflows/{workflow_name}/push",
            WorkflowPushRequest(
                workflow_dag=workflow_dag,
                activate=is_active,
            ).model_dump(),
        )
        return WorkflowPushResponse.model_validate(response)

    def list_workflows(self) -> list[str]:
        """
        Returns a list of workflow ids of all pushed workflows.

        Returns:
            A list of workflow ids.
        """
        result = self._get(f"{API_V1_STR}/workflows/list")
        return [Workflow.model_validate(r) for r in result]

    def trigger_workflow(
        self, workflow_name: str, payload: JsonValue | None
    ) -> WorkflowTriggerResponse:
        """
        Trigger a workflow with the given name.

        Args:
            workflow_name: The workflow name.
            payload: The payload to send to the workflow.
        """
        response = self._post(
            f"{API_V1_STR}/workflows/{workflow_name}/trigger", payload
        )
        return WorkflowTriggerResponse.model_validate(response)

    def activate_workflow(self, workflow_name: str) -> None:
        """
        Activates the workflow with the given name.

        Args:
            workflow_name: The workflow name.
        """
        self._post(
            f"{API_V1_STR}/workflows/activate", json={"workflow_name": workflow_name}
        )

    def deactivate_workflow(self, workflow_name: str) -> None:
        """
        Deactivates the workflow with the given name.

        Args:
            workflow_name: The workflow name.
        """
        self._post(
            f"{API_V1_STR}/workflows/deactivate",
            json={"workflow_name": workflow_name},
        )

    ########################################################
    # Python Action
    ########################################################

    def push_action(self, python_action: PythonAction) -> None:
        """
        Pushes the action to the server.

        Args:
            python_action: The action to push.
        """
        self._post(f"{API_V1_STR}/actions/push", python_action.model_dump())

    def get_action(self, action_type: str) -> PythonAction | None:
        """
        Returns the action with the given type if it exists.

        Args:
            action_type: The action type.

        Returns:
            The action with the given type if it exists, otherwise None.
        """
        result = self._get(f"{API_V1_STR}/actions/{action_type}")
        return PythonAction.model_validate(result) if result else None

    def list_actions(self) -> list[ActionMetadata]:
        """
        Returns a list of action names.

        Returns:
            A list of action names.
        """

        result = self._get(f"{API_V1_STR}/actions")
        return [ActionMetadata.model_validate(r) for r in result]

    def delete_action(self, action_type: str) -> None:
        """
        Deletes the action with the given type.

        Args:
            action_type: The action type.
        """
        self._delete(f"{API_V1_STR}/actions/{action_type}")

    ########################################################
    # Secrets
    ########################################################

    def set_secret(self, secret: Secret) -> None:
        """
        Sets the secret. If the secret already exists, it will be overwritten.
        Otherwise, a new secret will be created.

        Args:
            secret: The secret to set.
        """
        self._post(f"{API_V1_STR}/secrets/set", secret.model_dump())

    def list_secrets(self) -> list[SecretMetadata]:
        """
        Returns a list of secret names.

        Returns:
            A list of secret names.
        """
        result = self._get(f"{API_V1_STR}/secrets/list")
        return [SecretMetadata.model_validate(r) for r in result]

    def delete_secret(self, secret_id: str) -> None:
        """
        Deletes the secret with the given id.

        Args:
            secret_id: The secret id
        """
        self._delete(f"{API_V1_STR}/secrets/delete", {"secret_id": secret_id})

    ########################################################
    # Webhook
    ########################################################

    def trigger_webhook(
        self, webhook_id: str, webhook_secret: str, payload: JsonValue | None
    ) -> WorkflowTriggerResponse:
        """
        Triggers the webhook with the given id.

        Args:
            webhook_id: The webhook id.
            webhook_secret: The webhook secret.
            payload: The payload to send to the webhook.
        """
        response = self._request(
            "POST",
            f"/webhooks/{webhook_id}",
            json=payload,
            webhook_secret=webhook_secret,
        )
        return WorkflowTriggerResponse.model_validate(response)



from pydantic import BaseModel

from admyral.utils.singleton import Singleton


class SecretRegistry(metaclass=Singleton):
    _secrets: dict[str, BaseModel] = {}

    @classmethod
    def register_secret(cls, secret_type: str, secret: BaseModel) -> None:
        if secret_type in cls._secrets:
            raise ValueError(f"Secret with type '{secret_type}' already exists")
        cls._secrets[secret_type] = secret

    @classmethod
    def get_secret_schemas(cls) -> dict[str, list[str]]:
        return {
            secret_type: secret.model_json_schema()["properties"].keys()
            for secret_type, secret in cls._secrets.items()
        }

    @classmethod
    def is_registered(cls, secret_type: str) -> bool:
        return secret_type in cls._secrets

    @classmethod
    def validate_schema(cls, secret_type: str, secret: dict[str, str]) -> None:
        if not cls.is_registered(secret_type):
            raise ValueError(f"Secret type '{secret_type}' is not registered")
        secret_model = cls._secrets[secret_type]
        secret_model.model_validate(secret)


# needs to be defined after SecretRegistry, such that SecretRegistry is fully defined
# before any action is registered
from admyral.actions import *  # noqa: E402,F403



import click

from admyral.cli.cli import cli
from admyral.client import AdmyralClient
from admyral.compiler.action_parser import parse_action
from admyral.utils.telemetry import capture


@cli.group()
def action() -> None:
    """Action Management"""


# TODO: rework arguments + recursive dependencies collection (constants, functions, etc.) + automatically determine action_type
# TODO: instead of providing a file, we should search for the action automatically
@action.command(
    "push",
    help="Push a Python action to Admyral",
)
@click.argument("action_type", type=str)
@click.option("--action", "-a", type=str, help="Path to actions file", required=True)
@click.pass_context
def push(ctx: click.Context, action_type: str, action: str) -> None:
    """Push an action to Admyral"""
    capture(event_name="action:push")
    client: AdmyralClient = ctx.obj

    with open(action, "r") as f:
        action_str = f.read()

    python_action = parse_action(action_str, action_type)

    try:
        client.push_action(python_action)
    except Exception as e:
        click.echo("Failed to push action.")
        click.echo(f"Error: {e}")
        return

    click.echo(f"Action {action_type} pushed successfully.")


@action.command(
    "list",
    help="List all pushed custom actions",
)
@click.pass_context
def list(ctx: click.Context) -> None:
    """List all pushed custom actions"""
    capture(event_name="action:list")
    client: AdmyralClient = ctx.obj
    actions = client.list_actions()
    click.echo("Pushed actions:")
    for action in actions:
        click.echo(action.action_type)


@action.command("delete", help="Delete a pushed custom action")
@click.argument("action_type", type=str)
@click.pass_context
def delete(ctx: click.Context, action_type: str) -> None:
    """Delete a pushed custom action"""
    capture(event_name="action:delete")
    client: AdmyralClient = ctx.obj
    client.delete_action(action_type)
    click.echo(f"Action {action_type} deleted successfully.")



from typing import Annotated
import boto3
import boto3.session

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.utils.temporalio import serialize_datetime_in_json
from admyral.actions.integrations.shared.aws import AWSSecret


# TODO: add documentation for cross-account role setup
@action(
    display_name="List Vulnerabilities",
    display_namespace="Amazon Inspector2",
    description="List Amazon Inspector2 vulnerabilities",
    secrets_placeholders=["AWS_SECRET"],
)
def list_amazon_inspector2_vulnerabilities(
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time of the findings as ISO-8601 formatted string (inclusive)",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time of the findings as ISO-8601 formatted string (inclusive)",
        ),
    ] = None,
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of findings to return.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    aws_secret = ctx.get().secrets.get("AWS_SECRET")
    aws_secret = AWSSecret.model_validate(aws_secret)

    session = boto3.session.Session(
        aws_access_key_id=aws_secret.aws_access_key_id,
        aws_secret_access_key=aws_secret.aws_secret_access_key,
    )
    inspector_client = session.client("inspector2")

    filter = {}
    if start_time or end_time:
        date_filter = {}
        if start_time:
            date_filter["startInclusive"] = start_time
        if end_time:
            date_filter["endInclusive"] = end_time
        filter["firstObservedAt"] = [date_filter]

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/inspector2/client/list_findings.html
    paginator = inspector_client.get_paginator("list_findings")
    findings = []
    for page in paginator.paginate(
        filterCriteria=filter, PaginationConfig={"PageSize": 100}
    ):
        findings.extend(page.get("findings", []))
        if len(findings) >= limit:
            findings = findings[:limit]
            break

    findings = findings[:limit]

    return serialize_datetime_in_json(findings[:limit])



from typing import Annotated
from sqlalchemy import create_engine, text
from pydantic import BaseModel
from datetime import datetime

from admyral.action import action, ArgumentMetadata
from admyral.typings import JsonValue
from admyral.context import ctx
from admyral.secret.secret import register_secret


@register_secret(secret_type="Database")
class DatabaseSecret(BaseModel):
    uri: str


def _handle_datetime_in_row(row: tuple) -> tuple:
    return tuple(
        item.isoformat().replace("+00:00", "Z") if isinstance(item, datetime) else item
        for item in row
    )


@action(
    display_name="Run SQL Query",
    display_namespace="Database",
    description="Run a SQL query on a database. Supported databases: PostgreSQL, MySQL, and MariaDB.",
    secrets_placeholders=["DB_URI"],
)
def run_sql_query(
    sql_query: Annotated[
        str,
        ArgumentMetadata(
            display_name="SQL Query", description="The query to run on the database"
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("DB_URI")
    secret = DatabaseSecret.model_validate(secret)

    db_uri = secret.uri
    if db_uri.startswith("mysql://"):
        db_uri = db_uri.replace("mysql://", "mysql+pymysql://")
    if db_uri.startswith("postgres://"):
        db_uri = db_uri.replace("postgres://", "postgresql://")

    with create_engine(db_uri).connect() as connection:
        result = connection.execute(text(sql_query))
        if not result.returns_rows:
            return
        columns = result.keys()
        return [
            dict(zip(columns, _handle_datetime_in_row(row)))
            for row in result.fetchall()
        ]



from admyral.action import action
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.actions.integrations.shared.ms_graph import (
    ms_graph_list_managed_devices,
    AzureSecret,
)


@action(
    display_name="List Managed Devices",
    display_namespace="Microsoft Intune",
    description="List managed devices from Microsoft Intune",
    secrets_placeholders=["AZURE_SECRET"],
)
def list_ms_intune_managed_devices() -> list[dict[str, JsonValue]]:
    secret = ctx.get().secrets.get("AZURE_SECRET")
    secret = AzureSecret.model_validate(secret)
    return ms_graph_list_managed_devices(secret)


@action(
    display_name="List Unencrypted Managed Devices",
    display_namespace="Microsoft Intune",
    description="List unencrypted managed devices from Microsoft Intune",
    secrets_placeholders=["AZURE_SECRET"],
)
def list_ms_intune_unencrypted_managed_devices() -> list[dict[str, JsonValue]]:
    devices = list_ms_intune_managed_devices()
    return [device for device in devices if not device["isEncrypted"]]



from admyral.actions.integrations.compliance.retool import (
    list_retool_inactive_users,
    list_groups_per_user,
    list_groups_and_apps_per_user,
    list_used_groups_and_apps_per_user,
)
from admyral.actions.integrations.compliance.one_password import (
    list_1password_audit_events,
)
from admyral.actions.integrations.compliance.github import (
    search_github_enterprise_audit_logs,
    list_github_merged_pull_requests,
    list_github_commit_history_for_pull_request,
    list_github_review_history_for_pull_request,
    compare_two_github_commits,
    list_github_merged_pull_requests_without_approval,
    list_github_issue_comments,
)
from admyral.actions.integrations.compliance.google_drive import (
    list_google_docs_revisions,
    list_google_drive_files_with_link_sharing_enabled,
)
from admyral.actions.integrations.compliance.ms_intune import (
    list_ms_intune_managed_devices,
    list_ms_intune_unencrypted_managed_devices,
)
from admyral.actions.integrations.compliance.kandji import (
    list_kandji_devices,
    get_kandji_device_details,
    list_kandji_unencrypted_devices,
    get_kandji_device_apps,
    get_kandji_application_firewall,
    get_kandji_desktop_and_screensaver,
    get_kandji_library_item_statuses,
)
from admyral.actions.integrations.compliance.zendesk import list_zendesk_users

__all__ = [
    "list_retool_inactive_users",
    "list_groups_per_user",
    "list_1password_audit_events",
    "search_github_enterprise_audit_logs",
    "list_github_merged_pull_requests",
    "list_github_commit_history_for_pull_request",
    "list_github_review_history_for_pull_request",
    "compare_two_github_commits",
    "list_github_merged_pull_requests_without_approval",
    "list_google_docs_revisions",
    "list_ms_intune_managed_devices",
    "list_ms_intune_unencrypted_managed_devices",
    "list_github_issue_comments",
    "list_kandji_devices",
    "get_kandji_device_details",
    "list_kandji_unencrypted_devices",
    "get_kandji_device_apps",
    "list_zendesk_users",
    "get_kandji_application_firewall",
    "get_kandji_desktop_and_screensaver",
    "get_kandji_library_item_statuses",
    "list_groups_and_apps_per_user",
    "list_used_groups_and_apps_per_user",
    "list_google_drive_files_with_link_sharing_enabled",
]



from admyral.db.admyral_store import AdmyralStore
from admyral.workers.workers_client import WorkersClient
from admyral.secret.secrets_manager import SecretsManager, secrets_manager_factory
from admyral.config.config import CONFIG


_admyral_store: AdmyralStore = None
_workers_client: WorkersClient = None
_secrets_manager: SecretsManager = None


async def setup_admyral_store() -> None:
    global _admyral_store
    _admyral_store = await AdmyralStore.create_store()


def get_admyral_store() -> AdmyralStore:
    global _admyral_store
    if not _admyral_store:
        raise RuntimeError("Admyral Store not initialized.")
    return _admyral_store


async def setup_workers_client() -> None:
    global _workers_client
    config = CONFIG
    _workers_client = await WorkersClient.connect(
        get_admyral_store(), config.temporal_host
    )


def get_workers_client() -> WorkersClient:
    global _workers_client
    if not _workers_client:
        raise RuntimeError("Workers Client not initialized.")
    return _workers_client


async def setup_secrets_manager() -> None:
    global _secrets_manager
    if not _admyral_store:
        raise RuntimeError(
            "Admyral Store is not initialized. Admyral Store must be initialized before Secrets Manager."
        )
    _secrets_manager = secrets_manager_factory(_admyral_store)


def get_secrets_manager() -> SecretsManager:
    global _secrets_manager
    if not _secrets_manager:
        raise RuntimeError("Secrets Manager not initialized.")
    return _secrets_manager


async def setup_dependencies():
    await setup_admyral_store()
    await setup_secrets_manager()
    await setup_workers_client()



from typing import Annotated
from openai import OpenAI
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.secret.secret import register_secret


@register_secret(secret_type="OpenAI")
class OpenAISecret(BaseModel):
    api_key: str


@action(
    display_name="Chat Completion",
    display_namespace="OpenAI",
    description="Use advanced AI models from OpenAI to perform complex tasks, such as categorization, "
    "analysis, summarization, or decision support.",
    secrets_placeholders=["OPENAI_SECRET"],
)
def openai_chat_completion(
    model: Annotated[
        str,
        ArgumentMetadata(
            display_name="Model",
            description="The model to use.",
        ),
    ],
    prompt: Annotated[
        str,
        ArgumentMetadata(
            display_name="Prompt",
            description="The prompt to use for the chat completion.",
        ),
    ],
    top_p: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Top P",
            description="Value between 0 and 1. An alternative to sampling with temperature, "
            "called nucleus sampling, where the model considers the results of the tokens "
            "with Top P probability mass. So 0.1 means only the tokens comprising the top 10% "
            "probability mass are considered. We generally recommend altering this or temperature "
            "but not both.",
        ),
    ] = None,
    temperature: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Temperature",
            description="What sampling temperature to use, between 0 and 2. Higher values like 0.8 "
            "will make the output more random, while lower values like 0.2 will make it more focused "
            "and deterministic. We generally recommend altering this or Top P but not both.",
        ),
    ] = None,
    stop_tokens: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Stop Tokens",
            description="A list of tokens at which to stop the completion.",
        ),
    ] = None,
) -> str:
    # https://platform.openai.com/docs/api-reference/chat/create
    # TODO: error handling
    secret = ctx.get().secrets.get("OPENAI_SECRET")
    openai_secret = OpenAISecret.model_validate(secret)

    model_params = {}
    if top_p is not None:
        model_params["top_p"] = top_p
    if temperature is not None:
        model_params["temperature"] = temperature
    if stop_tokens is not None and not model.startswith("o1"):
        model_params["stop"] = stop_tokens

    client = OpenAI(api_key=openai_secret.api_key)
    chat_completion = client.chat.completions.create(
        messages=[{"role": "user", "content": prompt}],
        model=model,
        **model_params,
    )

    return chat_completion.choices[0].message.content



"""add secret_type to to secrets table

Revision ID: 2f91aed0b218
Revises: fa67e2abfd00
Create Date: 2024-11-03 23:38:43.477126

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "2f91aed0b218"
down_revision: Union[str, None] = "fa67e2abfd00"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("secrets", sa.Column("secret_type", sa.TEXT(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("secrets", "secret_type")
    # ### end Alembic commands ###



from admyral.actions.integrations.vulnerability_management.amazon_inspector2 import (
    list_amazon_inspector2_vulnerabilities,
)
from admyral.actions.integrations.vulnerability_management.snyk import (
    list_snyk_vulnerabilities,
)

__all__ = [
    "list_amazon_inspector2_vulnerabilities",
    "list_snyk_vulnerabilities",
]



import re
import json

from admyral.typings import JsonValue
from admyral.exceptions import AdmyralFailureError


REFERENCE_REGEX = re.compile(r"{{((?!}}).)*}}")


PATTERN1 = r"\[\"((?!\"\]).)+\"\]"
OBJECT_ACCESS_REGEX = re.compile(PATTERN1)

PATTERN2 = r"\[\'((?!\'\]).)+\'\]"
OBJECT_ACCESS_2_REGEX = re.compile(PATTERN2)

PATTERN3 = r"\[((?!\])-?[0-9])+\]"
INDEX_ACCESS_REGEX = re.compile(PATTERN3)

ACCESS_PATH_REGEX = re.compile(r"\[((?!\]).)*\]")


def _resolve_access_path(action_outputs: dict, input: str) -> JsonValue:
    input = input.strip()
    if not input.startswith("{{") or not input.endswith("}}"):
        # we have a JSON-serialized constant as input
        return json.loads(input)

    access_path = input.lstrip("{{").rstrip("}}").strip()
    if len(access_path) == 0:
        raise AdmyralFailureError(message="Invalid reference: Access path is empty.")
    current_value = action_outputs

    # access the base variable and check if it exists
    variable = access_path
    access_path_start = access_path.find("[")
    if access_path_start != -1:
        variable = access_path[:access_path_start]
    current_value = current_value.get(variable)
    if current_value is None:
        return None

    for key in ACCESS_PATH_REGEX.finditer(access_path):
        raw_value = key.group()

        if OBJECT_ACCESS_REGEX.match(raw_value) or OBJECT_ACCESS_2_REGEX.match(
            raw_value
        ):
            # Case - Object Access: ['key'] or ["key"]
            raw_value = raw_value[2:-2]
            if not isinstance(current_value, dict):
                raise AdmyralFailureError(
                    message=f"Invalid access path: {access_path}. Expected a dictionary, got {type(current_value).__name__}."
                )
            if raw_value not in current_value:
                raise AdmyralFailureError(
                    message=f"Invalid access path: {access_path}. Key '{raw_value}' not found."
                )
            current_value = current_value[raw_value]
        elif INDEX_ACCESS_REGEX.match(raw_value):
            # Case - Array Access: [index]
            try:
                raw_value = int(raw_value[1:-1])
            except ValueError:
                raise AdmyralFailureError(
                    message=f"Invalid access path: {access_path}. Expected an integer, got {raw_value}."
                )
            if not isinstance(current_value, list):
                raise AdmyralFailureError(
                    message=f"Invalid access path: {access_path}. Expected a list, got {type(current_value).__name__}."
                )
            if raw_value >= len(current_value) or raw_value < -len(current_value):
                raise AdmyralFailureError(
                    message=f"Invalid access path: {access_path}. Index {raw_value} out of bounds."
                )
            current_value = current_value[raw_value]
        else:
            raise AdmyralFailureError(
                message=f"Invalid access path segment: {access_path}. Must be either a string or integer."
            )

    return current_value


def evaluate_references(value: JsonValue, execution_state: dict) -> JsonValue:
    if value is None or isinstance(value, (bool, int, float)):
        return value

    if isinstance(value, str):
        reference_matches = [match.group() for match in REFERENCE_REGEX.finditer(value)]

        if (
            value.startswith("{{")
            and value.endswith("}}")
            and len(reference_matches) == 1
        ):
            # We have something like: "{{ reference }}"
            return _resolve_access_path(execution_state, value)

        for reference in reference_matches:
            resolved_ref = _resolve_access_path(execution_state, reference)
            resolved_ref_str = str(resolved_ref) if resolved_ref is not None else "null"
            value = value.replace(reference, resolved_ref_str)
        return value

    if isinstance(value, dict):
        out = {}
        for key, val in value.items():
            key = evaluate_references(key, execution_state)
            value = evaluate_references(val, execution_state)
            out[key] = value
        return out

    if isinstance(value, list):
        return [evaluate_references(val, execution_state) for val in value]

    raise AdmyralFailureError(message=f"Unsupported type: {type(value).__name__}")



from pydantic import BaseModel

from admyral.secret.secret import register_secret


@register_secret(secret_type="AWS")
class AWSSecret(BaseModel):
    aws_access_key_id: str
    aws_secret_access_key: str



from admyral.actions.integrations.enrich.alienvault_otx import (
    alienvault_otx_analyze_domain,
)
from admyral.actions.integrations.enrich.greynoise import grey_noise_ip_lookup
from admyral.actions.integrations.enrich.virus_total import (
    virus_total_analyze_hash,
    virus_total_analyze_domain,
    virus_total_analyze_ip,
    virus_total_analyze_url,
)
from admyral.actions.integrations.enrich.abuseipdb import abuseipdb_analyze_ip
from admyral.actions.integrations.enrich.leakcheck import (
    leakcheck_v2_lookup,
    leakcheck_public_lookup,
)

__all__ = [
    "alienvault_otx_analyze_domain",
    "grey_noise_ip_lookup",
    "virus_total_analyze_hash",
    "virus_total_analyze_domain",
    "virus_total_analyze_ip",
    "virus_total_analyze_url",
    "abuseipdb_analyze_ip",
    "leakcheck_v2_lookup",
    "leakcheck_public_lookup",
]



from typing import Any


def throw_if_not_allowed_return_type(obj: Any) -> None:
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return
    elif isinstance(obj, list):
        for item in obj:
            throw_if_not_allowed_return_type(item)
    elif isinstance(obj, dict):
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ValueError(
                    f"Key of type {type(key)} is not allowed in the JSON object. The key must be a string."
                )
            throw_if_not_allowed_return_type(value)
    else:
        raise ValueError(
            f"Object of type {type(obj)} is not allowed in the return statement. The type must be JSON."
        )



import click

from admyral.client import AdmyralClient
from admyral.config.config import CONFIG


@click.group()
@click.pass_context
def cli(ctx):
    """Admyral CLI

    Admyral is an open-source Python SDK enabling Security Engineers to build complex workflow automations using Python — blazingly fast.
    """
    ctx.obj = AdmyralClient(base_url=CONFIG.cli_target, api_key=CONFIG.api_key)



"""init

Revision ID: 014da66fd11f
Revises:
Create Date: 2024-08-17 23:06:22.463683

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "014da66fd11f"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "pip_lockfile_cache",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("hash", sa.TEXT(), nullable=False),
        sa.Column("lockfile", sa.TEXT(), nullable=False),
        sa.Column("expiration_time", sa.TIMESTAMP(), nullable=False),
        sa.PrimaryKeyConstraint("hash"),
    )
    op.create_table(
        "python_actions",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("user_id", sa.TEXT(), nullable=False),
        sa.Column("action_type", sa.TEXT(), nullable=False),
        sa.Column("import_statements", sa.TEXT(), nullable=False),
        sa.Column("code", sa.TEXT(), nullable=False),
        sa.Column("display_name", sa.TEXT(), nullable=False),
        sa.Column("display_namespace", sa.TEXT(), nullable=False),
        sa.Column("description", sa.TEXT(), nullable=True),
        sa.Column("secrets_placeholders", sa.TEXT(), nullable=True),
        sa.Column("requirements", sa.TEXT(), nullable=True),
        sa.Column("arguments", sa.JSON(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", "action_type"),
    )
    op.create_table(
        "secrets",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("secret_id", sa.TEXT(), nullable=False),
        sa.Column("user_id", sa.TEXT(), nullable=False),
        sa.Column("encrypted_secret", sa.TEXT(), nullable=True),
        sa.Column("schema_json_serialized", sa.TEXT(), nullable=False),
        sa.PrimaryKeyConstraint("secret_id"),
    )
    op.create_index(op.f("ix_secrets_user_id"), "secrets", ["user_id"], unique=False)
    op.create_table(
        "workflows",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("workflow_id", sa.TEXT(), nullable=False),
        sa.Column("user_id", sa.TEXT(), nullable=False),
        sa.Column("workflow_name", sa.TEXT(), nullable=False),
        sa.Column("workflow_dag", sa.JSON(), nullable=False),
        sa.Column("is_active", sa.BOOLEAN(), nullable=False),
        sa.PrimaryKeyConstraint("workflow_id"),
        sa.UniqueConstraint("user_id", "workflow_name", name="unique_workflow_name"),
    )
    op.create_index(
        op.f("ix_workflows_user_id"), "workflows", ["user_id"], unique=False
    )
    op.create_index(
        op.f("ix_workflows_workflow_name"), "workflows", ["workflow_name"], unique=False
    )
    op.create_table(
        "workflow_runs",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("run_id", sa.TEXT(), nullable=False),
        sa.Column("user_id", sa.TEXT(), nullable=False),
        sa.Column("workflow_id", sa.TEXT(), nullable=False),
        sa.Column("source_name", sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(
            ["workflow_id"], ["workflows.workflow_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("run_id"),
    )
    op.create_table(
        "workflow_schedules",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("schedule_id", sa.TEXT(), nullable=False),
        sa.Column("user_id", sa.TEXT(), nullable=False),
        sa.Column("workflow_id", sa.TEXT(), nullable=False),
        sa.Column("cron", sa.TEXT(), nullable=True),
        sa.Column("interval_seconds", sa.INTEGER(), nullable=True),
        sa.Column("interval_minutes", sa.INTEGER(), nullable=True),
        sa.Column("interval_hours", sa.INTEGER(), nullable=True),
        sa.Column("interval_days", sa.INTEGER(), nullable=True),
        sa.Column("default_args", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(
            ["workflow_id"], ["workflows.workflow_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("schedule_id"),
    )
    op.create_index(
        op.f("ix_workflow_schedules_user_id"),
        "workflow_schedules",
        ["user_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_workflow_schedules_workflow_id"),
        "workflow_schedules",
        ["workflow_id"],
        unique=False,
    )
    op.create_table(
        "workflow_webhooks",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("webhook_id", sa.TEXT(), nullable=False),
        sa.Column("workflow_id", sa.TEXT(), nullable=False),
        sa.Column("webhook_secret", sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(
            ["workflow_id"], ["workflows.workflow_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("webhook_id"),
    )
    op.create_index(
        op.f("ix_workflow_webhooks_workflow_id"),
        "workflow_webhooks",
        ["workflow_id"],
        unique=False,
    )
    op.create_table(
        "workflow_run_steps",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("step_id", sa.TEXT(), nullable=False),
        sa.Column("run_id", sa.TEXT(), nullable=False),
        sa.Column("prev_step_id", sa.TEXT(), nullable=True),
        sa.Column("action_type", sa.TEXT(), nullable=False),
        sa.Column("input_args", sa.JSON(), nullable=True),
        sa.Column("logs", sa.TEXT(), nullable=True),
        sa.Column("result", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(
            ["run_id"], ["workflow_runs.run_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("step_id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("workflow_run_steps")
    op.drop_index(
        op.f("ix_workflow_webhooks_workflow_id"), table_name="workflow_webhooks"
    )
    op.drop_table("workflow_webhooks")
    op.drop_index(
        op.f("ix_workflow_schedules_workflow_id"), table_name="workflow_schedules"
    )
    op.drop_index(
        op.f("ix_workflow_schedules_user_id"), table_name="workflow_schedules"
    )
    op.drop_table("workflow_schedules")
    op.drop_table("workflow_runs")
    op.drop_index(op.f("ix_workflows_workflow_name"), table_name="workflows")
    op.drop_index(op.f("ix_workflows_user_id"), table_name="workflows")
    op.drop_table("workflows")
    op.drop_index(op.f("ix_secrets_user_id"), table_name="secrets")
    op.drop_table("secrets")
    op.drop_table("python_actions")
    op.drop_table("pip_lockfile_cache")
    # ### end Alembic commands ###



from sqlmodel import (
    Field,
    SQLModel,
    UniqueConstraint,
    ForeignKeyConstraint,
    Relationship,
)
from sqlalchemy import TEXT, TIMESTAMP, BOOLEAN, Column
from datetime import datetime

from admyral.models.auth import User
from admyral.db.schemas.api_key_schemas import ApiKeySchema
from admyral.db.schemas.workflow_schemas import WorkflowSchema
from admyral.db.schemas.secrets_schemas import SecretsSchema
from admyral.db.schemas.base_schemas import BaseSchema


class UserSchema(BaseSchema, table=True):
    """
    Schema for Users

    Needs to match the NextAuth schema:
    https://authjs.dev/getting-started/adapters/prisma
    """

    __tablename__ = "User"
    __table_args__ = (UniqueConstraint("email"),)

    id: str = Field(sa_type=TEXT(), primary_key=True)
    name: str | None = Field(sa_type=TEXT(), nullable=True)
    email: str = Field(sa_type=TEXT())
    email_verified: datetime | None = Field(sa_type=TIMESTAMP(), nullable=True)
    image: str | None = Field(sa_type=TEXT(), nullable=True)

    # relationship children
    accounts: list["AccountSchema"] = Relationship(
        back_populates="user", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    sessions: list["SessionSchema"] = Relationship(
        back_populates="user", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    authenticators: list["AuthenticatorSchema"] = Relationship(
        back_populates="user", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    workflows: list[WorkflowSchema] = Relationship(
        back_populates="user", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    api_keys: list[ApiKeySchema] = Relationship(
        back_populates="user", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    secrets: list[SecretsSchema] = Relationship(
        back_populates="user", sa_relationship_kwargs=dict(cascade="all, delete")
    )

    def to_model(self) -> User:
        return User.model_validate(
            {
                "id": self.id,
                "name": self.name,
                "email": self.email,
                "email_verified": self.email_verified,
                "image": self.image,
                "created_at": self.created_at,
                "updated_at": self.updated_at,
            }
        )


class AccountSchema(BaseSchema, table=True):
    """
    Schema for Accounts

    Needs to match the NextAuth schema:
    https://authjs.dev/getting-started/adapters/prisma
    """

    __tablename__ = "Account"
    __table_args__ = (
        UniqueConstraint("provider", "providerAccountId"),
        ForeignKeyConstraint(
            ["userId"],
            ["User.id"],
            ondelete="CASCADE",
        ),
    )

    user_id: str = Field(sa_column=Column("userId", TEXT(), nullable=False))
    type: str = Field(sa_type=TEXT())
    provider: str = Field(sa_type=TEXT(), primary_key=True)
    provider_account_id: str = Field(
        sa_column=Column("providerAccountId", TEXT(), nullable=False, primary_key=True)
    )
    refresh_token: str | None = Field(sa_type=TEXT(), nullable=True)
    access_token: str | None = Field(sa_type=TEXT(), nullable=True)
    expires_at: int | None = Field(nullable=True)
    token_type: str | None = Field(sa_type=TEXT(), nullable=True)
    scope: str | None = Field(sa_type=TEXT(), nullable=True)
    id_token: str | None = Field(sa_type=TEXT(), nullable=True)
    session_state: str | None = Field(sa_type=TEXT(), nullable=True)

    # relationship parents
    user: "UserSchema" = Relationship(back_populates="accounts")


class SessionSchema(BaseSchema, table=True):
    """
    Schema for Session

    Needs to match the NextAuth schema:
    https://authjs.dev/getting-started/adapters/prisma
    """

    __tablename__ = "Session"
    __table_args__ = (
        UniqueConstraint("sessionToken"),
        ForeignKeyConstraint(
            ["userId"],
            ["User.id"],
            ondelete="CASCADE",
        ),
    )

    id: str = Field(sa_type=TEXT(), primary_key=True)
    session_token: str = Field(sa_column=Column("sessionToken", TEXT(), nullable=False))
    user_id: str = Field(sa_column=Column("userId", TEXT(), nullable=False))
    expires: datetime

    # relationship parents
    user: "UserSchema" = Relationship(back_populates="sessions")


class VerificationTokenSchema(SQLModel, table=True):
    """
    Schema for VerificationToken

    Needs to match the NextAuth schema:
    https://authjs.dev/getting-started/adapters/prisma
    """

    __tablename__ = "VerificationToken"
    __table_args__ = (UniqueConstraint("identifier", "token"),)

    identifier: str = Field(sa_type=TEXT(), primary_key=True)
    token: str = Field(sa_type=TEXT(), primary_key=True)
    expires: datetime


class AuthenticatorSchema(SQLModel, table=True):
    """
    Schema for Authenticator

    Needs to match the NextAuth schema:
    https://authjs.dev/getting-started/adapters/prisma
    """

    __tablename__ = "Authenticator"
    __table_args__ = (
        UniqueConstraint("credentialID"),
        ForeignKeyConstraint(
            ["userId"],
            ["User.id"],
            ondelete="CASCADE",
        ),
    )

    credential_id: str = Field(
        sa_column=Column("credentialID", TEXT(), primary_key=True)
    )
    user_id: str = Field(sa_column=Column("userId", TEXT(), primary_key=True))
    provider_account_id: str = Field(
        sa_column=Column("providerAccountId", TEXT(), nullable=False)
    )
    credential_public_key: str = Field(
        sa_column=Column("credentialPublicKey", TEXT(), nullable=False)
    )
    counter: int
    credential_device_type: str = Field(
        sa_column=Column("credentialDeviceType", TEXT(), nullable=False)
    )
    credentialBackedUp: bool = Field(
        sa_column=Column("credentialBackedUp", BOOLEAN(), nullable=False)
    )
    transports: str | None = Field(sa_type=TEXT(), nullable=True)

    # relationship parents
    user: "UserSchema" = Relationship(back_populates="authenticators")



from typing import Any
from multipledispatch import dispatch

from admyral.typings import JsonValue
from admyral.models import (
    UnaryOperator,
    BinaryOperator,
    IVisitor,
    ConditionExpression,
    ConstantConditionExpression,
    UnaryConditionExpression,
    BinaryConditionExpression,
    AndConditionExpression,
    OrConditionExpression,
    Condition,
    condition_validate,
)
from admyral.workers.references import evaluate_references


# TODO: error handling for incompatible types
class ConditionEvaluator(IVisitor):
    @dispatch(ConstantConditionExpression)
    def visit(self, expr: ConstantConditionExpression) -> Any:  # noqa F811
        return expr.value

    @dispatch(UnaryConditionExpression)
    def visit(self, expr: UnaryConditionExpression) -> Any:  # noqa F811
        match expr.op:
            case UnaryOperator.NOT:
                return not expr.expr.accept(self)
            case UnaryOperator.IS_NONE:
                return expr.expr.accept(self) is None
            case UnaryOperator.IS_NOT_NONE:
                return expr.expr.accept(self) is not None

    @dispatch(BinaryConditionExpression)
    def visit(self, expr: BinaryConditionExpression) -> Any:  # noqa F811
        lhs = expr.lhs.accept(self)
        rhs = expr.rhs.accept(self)
        match expr.op:
            case BinaryOperator.EQUALS:
                return lhs == rhs
            case BinaryOperator.NOT_EQUALS:
                return lhs != rhs
            case BinaryOperator.GREATER_THAN:
                return lhs > rhs
            case BinaryOperator.LESS_THAN:
                return lhs < rhs
            case BinaryOperator.GREATER_THAN_OR_EQUAL:
                return lhs >= rhs
            case BinaryOperator.LESS_THAN_OR_EQUAL:
                return lhs <= rhs
            case BinaryOperator.IN:
                return lhs in rhs
            case BinaryOperator.NOT_IN:
                return lhs not in rhs
            case _:
                raise ValueError(f"Invalid operator: {expr.op.value}")

    @dispatch(AndConditionExpression)
    def visit(self, expr: AndConditionExpression) -> Any:  # noqa F811
        return all([expr.accept(self) for expr in expr.and_expr])

    @dispatch(OrConditionExpression)
    def visit(self, expr: OrConditionExpression) -> Any:  # noqa F811
        return any([expr.accept(self) for expr in expr.or_expr])

    def evaluate(self, expr: ConditionExpression) -> bool:
        return bool(expr.accept(self))


class ConditionReferenceResolution(IVisitor):
    def __init__(self, execution_state: dict[str, JsonValue]) -> None:
        self.execution_state = execution_state

    @dispatch(ConstantConditionExpression)
    def visit(self, expr: ConstantConditionExpression) -> Any:  # noqa F811
        expr.value = evaluate_references(expr.value, self.execution_state)
        return expr

    @dispatch(UnaryConditionExpression)
    def visit(self, expr: UnaryConditionExpression) -> Any:  # noqa F811
        expr.expr.accept(self)
        return expr

    @dispatch(BinaryConditionExpression)
    def visit(self, expr: BinaryConditionExpression) -> Any:  # noqa F811
        expr.lhs.accept(self)
        expr.rhs.accept(self)
        return expr

    @dispatch(AndConditionExpression)
    def visit(self, expr: AndConditionExpression) -> Any:  # noqa F811
        expr.and_expr = [expr.accept(self) for expr in expr.and_expr]
        return expr

    @dispatch(OrConditionExpression)
    def visit(self, expr: OrConditionExpression) -> Any:  # noqa F811
        expr.or_expr = [expr.accept(self) for expr in expr.or_expr]
        return expr

    def resolve_references(self, expr: ConditionExpression) -> Condition:
        expr.accept(self)
        return expr


def execute_if_condition(condition_expr: JsonValue) -> bool:
    condition = condition_validate(condition_expr)
    return ConditionEvaluator().evaluate(condition)



import uvicorn


async def run_api(config: dict) -> None:
    config = uvicorn.Config(
        "admyral.server.server:app",
        host="0.0.0.0" if config.get("docker") else "127.0.0.1",
        port=8000,
    )
    server = uvicorn.Server(config)
    await server.serve()



"""transform workflow names

Revision ID: 2c07323c94bf
Revises: 552e7a8dec07
Create Date: 2024-11-02 14:33:05.196806

"""

from typing import Sequence, Union

from alembic import op
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "2c07323c94bf"
down_revision: Union[str, None] = "552e7a8dec07"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(
        "uq_Account_provider_providerAccountId",
        "Account",
        ["provider", "providerAccountId"],
    )
    op.create_unique_constraint(
        "uq_VerificationToken_identifier_token",
        "VerificationToken",
        ["identifier", "token"],
    )

    # transform workflow names from snake_case to Title Case
    op.execute("""
    UPDATE workflows
    SET 
        workflow_name = (
            SELECT string_agg(
                INITCAP(part),
                ' '
            )
            FROM unnest(string_to_array(workflow_name, '_')) part
        ),
        workflow_dag = jsonb_set(
            workflow_dag::jsonb,
            '{name}',
            to_jsonb(
                (
                    SELECT string_agg(
                        INITCAP(part),
                        ' '
                    )
                    FROM unnest(string_to_array(workflow_name, '_')) part
                )
            )
        )
    WHERE workflow_name LIKE '%_%';
    """)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    # transform workflow names to snake_case
    op.execute("""
    UPDATE workflows
    SET
        workflow_name = (
            SELECT string_agg(
                LOWER(part),
                '_'
            )
            FROM unnest(string_to_array(workflow_name, ' ')) part
        ),
        workflow_dag = jsonb_set(
            workflow_dag::jsonb,
            '{name}',
            to_jsonb(
                (
                    SELECT string_agg(
                        LOWER(part),
                        '_'
                    )
                    FROM unnest(string_to_array(workflow_name, ' ')) part
                )
            )
        )
    WHERE workflow_name LIKE '% %';
    """)

    op.drop_constraint(
        "uq_VerificationToken_identifier_token", "VerificationToken", type_="unique"
    )
    op.drop_constraint(
        "uq_Account_provider_providerAccountId", "Account", type_="unique"
    )
    # ### end Alembic commands ###



"""migrate secrets to secret_type

Revision ID: db5cf73c0869
Revises: 2f91aed0b218
Create Date: 2024-11-04 23:34:16.416482

"""

from typing import Sequence, Union

from alembic import op
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "db5cf73c0869"
down_revision: Union[str, None] = "2f91aed0b218"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    bind = op.get_bind()

    integrations = [
        ("anthropic", "Anthropic"),
        ("mistral ai", "Mistral AI"),
        ("openai", "OpenAI"),
        ["Jira", "Jira"],
        ["OpsGenie", "OpsGenie"],
        ["PagerDuty", "PagerDuty"],
        ["Intune", "Azure"],
        ["Defender", "Azure"],
        ["Azure", "Azure"],
        ("azure openai", "Azure OpenAI"),
        ["Wiz", "Wiz"],
        ["turbot", "Database"],
        ["steampipe", "Database"],
        ["slack", "Slack"],
        ["Github", "GitHub"],
        ["google", "Google"],
        ["drive", "Google"],
        ["kandji", "Kandji"],
        ["1Password", "1Password"],
        ["retool", "Retool"],
        ["zendesk", "Zendesk"],
        ["sentinelone", "SentinelOne"],
        ["abnormal", "Abnormal Security"],
        ["abuseipdb", "AbuseIPDB"],
        ["otx", "AlienVault OTX"],
        ["alienvault", "AlienVault OTX"],
        ["greynoise", "GreyNoise"],
        ["leakcheck", "LeakCheck"],
        ["virus", "VirusTotal"],
        ["okta", "Okta"],
        ["snyk", "Snyk"],
    ]

    for potential_name, secret_type in integrations:
        potential_name: str = potential_name.lower()
        bind.execute(
            sqlmodel.text(
                "UPDATE secrets SET secret_type = :new_type WHERE secret_type LIKE :pattern"
            ),
            {
                "new_type": secret_type,
                "pattern": f"%{potential_name}%",
            },
        )
        potential_name = potential_name.replace(" ", "_")
        bind.execute(
            sqlmodel.text(
                "UPDATE secrets SET secret_type = :new_type WHERE secret_type LIKE :pattern"
            ),
            {
                "new_type": secret_type,
                "pattern": f"%{potential_name}%",
            },
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    # since secret types have no use previously we can just
    # leave them as is
    pass

    # ### end Alembic commands ###



import asyncio
from asyncio.streams import StreamReader
from typing import Callable, Awaitable

from admyral.logger import get_logger

logger = get_logger(__name__)


async def _read_stream(stream: StreamReader, callback: Callable[[str], None]) -> None:
    try:
        while True:
            line = await stream.readline()
            if not line:
                break
            callback(line.decode("utf-8"))
    except ValueError:
        # ValueError: Separator is not found, and chunk exceed the limit
        # https://stackoverflow.com/questions/55457370/how-to-avoid-valueerror-separator-is-not-found-and-chunk-exceed-the-limit
        callback("Truncated log stream because chunks are too large to be read.")


async def run_subprocess_with_streamed_stdout_and_stderr(
    cmd: list[str],
    stdout_callback: Callable[[str], None],
    stderr_callback: Callable[[str], None],
    env: dict[str, str] | None = None,
) -> int:
    """
    Run a subprocess and stream stdout and stderr to the given callbacks.

    Args:
        cmd: The command to run.
        stdout_callback: The callback to call with each line of stdout.
        stderr_callback: The callback to call with each line of stderr.

    Returns:
        The exit code of the subprocess.
    """
    process = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE, env=env
    )

    async with asyncio.TaskGroup() as tg:
        tg.create_task(_read_stream(process.stdout, stdout_callback))
        tg.create_task(_read_stream(process.stderr, stderr_callback))

    await process.wait()

    return process.returncode


async def run_subprocess_with_log_flushing(
    cmd: list[str],
    append_logs_fn: Callable[[list[str]], Awaitable[None]],
    env: dict[str, str] | None = None,
) -> int:
    """
    Run a suboprocess and periodically call the append_logs_fn to flush the newest logs.

    Args:
        cmd: The command to run.
        append_logs_fn: The function to call with the newest logs to flush.

    Returns:
        The exit code of the subprocess.
    """

    # we first start a task which periodically flushes the lines from the buffer.
    # another task will be started which consumes the stdout and stderr stream from
    # the subprocess and appends the lines to the buffer.

    lines_buffer = []
    stop_flush_buffer_task = False

    async def flush_buffer():
        nonlocal lines_buffer
        nonlocal stop_flush_buffer_task

        while True:
            # consume the buffer before we append the logs. otherwise new logs
            # might have been appended while we were appending the logs.
            to_be_flushed = lines_buffer
            lines_buffer = []
            if len(to_be_flushed) > 0:
                await append_logs_fn(to_be_flushed)

            if stop_flush_buffer_task:
                # before we stop we need to check whether there are still logs to be flushed
                if len(lines_buffer) > 0:
                    await append_logs_fn(lines_buffer)
                break

            # sleep for 500ms
            await asyncio.sleep(0.5)

    flush_buffer_task = asyncio.create_task(flush_buffer())

    def append_new_line_to_buffer(new_line: str) -> None:
        nonlocal lines_buffer
        lines_buffer.append(new_line)

    process_return_code = await run_subprocess_with_streamed_stdout_and_stderr(
        cmd, append_new_line_to_buffer, append_new_line_to_buffer, env
    )

    logger.info(f"Subprocess exited with code {process_return_code}")

    # stop the flush_buffer task and wait until it completed
    stop_flush_buffer_task = True
    await flush_buffer_task

    if len(lines_buffer) > 0:
        await append_logs_fn(lines_buffer)

    return process_return_code



from typing import TypeVar, Callable, Any

from admyral.typings import JsonValue
from admyral.models import WorkflowDAG
from admyral.compiler.workflow_compiler import WorkflowCompiler


T = TypeVar("T", bound="Workflow")
F = TypeVar("F", bound=Callable[..., None])


class Webhook:
    def __init__(self, **default_args: dict[str, JsonValue]) -> None:
        super().__init__()
        self.default_args = default_args


class Schedule:
    def __init__(
        self,
        cron: str | None = None,
        interval_seconds: int | None = None,
        interval_minutes: int | None = None,
        interval_hours: int | None = None,
        interval_days: int | None = None,
        **default_args: dict[str, JsonValue],
    ) -> None:
        self.cron = cron
        self.interval_seconds = interval_seconds
        self.interval_minutes = interval_minutes
        self.interval_hours = interval_hours
        self.interval_days = interval_days
        self.default_args = default_args


class Workflow:
    def __init__(
        self,
        func: F,
        *,
        triggers: list[Webhook | Schedule] = [],
        description: str | None = None,
        controls: list[str] | None = None,
    ) -> None:
        self.func = func
        self.description = description
        self.triggers = triggers
        self.controls = controls

    def compile(self) -> WorkflowDAG:
        return WorkflowCompiler().compile(self)

    @property
    def name(self) -> str:
        return self.func.__name__

    def __call__(self, payload: dict[str, JsonValue]) -> Any:
        """Call the wrapped workflow function."""
        return self.func(payload)


# TODO: type overload + correct return type
def workflow(
    _func: F = None,
    *,
    description: str | None = None,
    triggers: list[Webhook | Schedule] = [],
    controls: list[str] | None = None,
) -> Workflow:
    """Decorator to create a workflow."""

    def inner(func: "F") -> Workflow:
        w = Workflow(
            func, description=description, triggers=triggers, controls=controls
        )
        w.__doc__ = func.__doc__
        return w

    return inner if _func is None else inner(_func)



from fastapi import APIRouter, status, Depends

from admyral.server.auth import authenticate
from admyral.models import AuthenticatedUser, UserProfile
from admyral.server.deps import get_admyral_store


router = APIRouter()


@router.get("", status_code=status.HTTP_200_OK)
async def get_user_profile(
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> UserProfile:
    return UserProfile(email=authenticated_user.email)


@router.delete("", status_code=status.HTTP_204_NO_CONTENT)
async def delete_user_profile(
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> None:
    await get_admyral_store().delete_user(authenticated_user.user_id)



from typing import Annotated
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="GreyNoise")
class GreyNoiseSecret(BaseModel):
    api_key: str


def get_grey_noise_client(secret: GreyNoiseSecret) -> Client:
    return Client(
        base_url="https://api.greynoise.io/v3",
        headers={
            "x-apikey": secret.api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Analyze IP Address",
    display_namespace="GreyNoise",
    description="Analyze an IP address using GreyNoise",
    secrets_placeholders=["GREY_NOISE_SECRET"],
)
def grey_noise_ip_lookup(
    ip_address: Annotated[
        str,
        ArgumentMetadata(
            display_name="IP Address", description="The IP address to analyze"
        ),
    ],
) -> JsonValue:
    # https://docs.greynoise.io/reference/get_v3-community-ip
    secret = ctx.get().secrets.get("GREY_NOISE_SECRET")
    secret = GreyNoiseSecret.model_validate(secret)

    with get_grey_noise_client(secret) as client:
        response = client.get(f"/community/{ip_address}")
        response.raise_for_status()
        return response.json()



from typing import Optional
import contextvars

from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.utils.future_executor import execute_future
from admyral.secret.secrets_access import Secrets
from admyral.config.config import CONFIG
from admyral.exceptions import NonRetryableActionError


class ExecutionContext:
    def __init__(
        self,
        workflow_id: str,
        run_id: str,
        action_type: str,
        user_id: str,
        step_id: Optional[str] = None,
        prev_step_id: Optional[str] = None,
        secrets: Secrets = Secrets.default(),
        _is_placeholder: bool = False,
    ) -> None:
        self.workflow_id = workflow_id
        self.run_id = run_id
        self.action_type = action_type
        self.step_id = step_id or ""
        self.prev_step_id = prev_step_id or ""
        self.user_id = user_id
        self.secrets = secrets
        self._is_placeholder = _is_placeholder

    @classmethod
    def placeholder(cls) -> "ExecutionContext":
        return ExecutionContext(
            workflow_id="", run_id="", action_type="", user_id=CONFIG.default_user_id
        )

    async def append_logs_async(self, lines: list[str]) -> None:
        if not self._is_placeholder:
            await SharedWorkerState.get_store().append_logs(
                self.step_id, self.run_id, self.action_type, self.prev_step_id, lines
            )

    def append_logs_sync(self, lines: list[str]) -> None:
        if not self._is_placeholder:
            execute_future(self.append_logs_async(lines))

    async def send_to_workflow_async(
        self, workflow_name: str, data: dict[str, str]
    ) -> None:
        if self._is_placeholder:
            return

        workflow = await SharedWorkerState.get_store().get_workflow_by_name(
            self.user_id, workflow_name
        )
        if not workflow:
            raise NonRetryableActionError(
                f'Failed to send to workflow "{workflow_name}". Workflow not found.'
            )

        await SharedWorkerState.get_workers_client().start_workflow(
            self.user_id, workflow, self.action_type, payload=data
        )

    def send_to_workflow_sync(self, workflow_name: str, data: dict[str, str]) -> None:
        if not self._is_placeholder:
            execute_future(self.send_to_workflow_async(workflow_name, data))


ctx = contextvars.ContextVar(
    "execution_context", default=ExecutionContext.placeholder()
)



from admyral.actions.integrations.edr.ms_defender import (
    list_ms_defender_for_endpoint_alerts,
)
from admyral.actions.integrations.edr.sentinel_one import list_sentinel_one_alerts

__all__ = [
    "list_ms_defender_for_endpoint_alerts",
    "list_sentinel_one_alerts",
]



import platform
from admyral.config.config import CONFIG
from admyral import __version__
from datetime import UTC, datetime
from requests import post


def _get_os() -> str:
    return platform.system()


def _get_detailed_platform() -> str:
    return platform.platform()


def _get_python_version() -> str:
    return platform.python_version()


def capture(event_name: str, properties: dict = {}) -> None:
    if CONFIG.telemetry_disabled or CONFIG.environment == "dev":
        return

    default_properties = {
        "admyral_version": __version__,
        "python_version": _get_python_version(),
        "user_os": _get_os(),
        "user_os_detail": _get_detailed_platform(),
    }

    post(
        url=f"{CONFIG.posthog_host}/capture",
        json={
            "api_key": CONFIG.posthog_api_key,
            "distinct_id": CONFIG.id,
            "event": f"{event_name}",
            "properties": default_properties | properties,
            "timestamp": str(datetime.now(UTC)),
        },
    )



from typing import Any
from datetime import datetime

from admyral.typings import JsonValue


def serialize_datetime_in_json(obj: Any) -> JsonValue:
    """
    Function to serialize datetime objects in JSON format.
    """
    if isinstance(obj, datetime):
        return f"{obj.isoformat()}Z"
    if isinstance(obj, list):
        return [serialize_datetime_in_json(item) for item in obj]
    if isinstance(obj, dict):
        return {key: serialize_datetime_in_json(value) for key, value in obj.items()}
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj
    raise ValueError(
        f"Object of type {type(obj)} is not a supported type. The type must be JSON-compatible."
    )



from typing import TYPE_CHECKING
from sqlmodel import (
    Field,
    Relationship,
    UniqueConstraint,
    JSON,
    TEXT,
    BOOLEAN,
    ForeignKeyConstraint,
)

from admyral.db.schemas.base_schemas import BaseSchema
from admyral.models import Workflow, WorkflowMetadata
from admyral.db.schemas.workflow_run_schemas import WorkflowRunSchema
from admyral.db.schemas.workflow_webhook_schemas import WorkflowWebhookSchema
from admyral.db.schemas.workflow_schedule_schemas import WorkflowScheduleSchema
from admyral.typings import JsonValue

if TYPE_CHECKING:
    from admyral.db.schemas.auth_schemas import UserSchema


class WorkflowSchema(BaseSchema, table=True):
    """
    Schema for Workflow
    """

    __tablename__ = "workflows"
    # workflow names must be unique per user!
    __table_args__ = (
        UniqueConstraint("user_id", "workflow_name", name="unique_workflow_name"),
        ForeignKeyConstraint(
            ["user_id"],
            ["User.id"],
            ondelete="CASCADE",
        ),
    )

    # primary keys
    workflow_id: str = Field(sa_type=TEXT(), primary_key=True)

    # foreign keys
    user_id: str = Field(
        sa_type=TEXT(), index=True
    )  # index for faster workflow listing

    # other fields
    workflow_name: str = Field(
        sa_type=TEXT(), index=True
    )  # index for faster workflow loading
    workflow_dag: JsonValue = Field(sa_type=JSON())

    # relationship parent
    user: "UserSchema" = Relationship(back_populates="workflows")

    # relationship children
    workflow_runs: list[WorkflowRunSchema] = Relationship(
        back_populates="workflow", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    webhooks: list[WorkflowWebhookSchema] = Relationship(
        back_populates="workflow", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    schedules: list[WorkflowScheduleSchema] = Relationship(
        back_populates="workflow", sa_relationship_kwargs=dict(cascade="all, delete")
    )
    is_active: bool = Field(sa_type=BOOLEAN(), default=False)

    def to_model(self, include_resources: bool = False) -> Workflow:
        # TODO: handle workflow runs, webhooks, schedules
        return Workflow.model_validate(
            {
                "workflow_id": self.workflow_id,
                "workflow_name": self.workflow_name,
                "workflow_dag": self.workflow_dag,
                "is_active": self.is_active,
            }
        )

    def to_metadata(self) -> WorkflowMetadata:
        return WorkflowMetadata(
            workflow_id=self.workflow_id,
            workflow_name=self.workflow_name,
            workflow_description=self.workflow_dag.get("description"),
            controls=self.workflow_dag.get("controls"),
            created_at=self.created_at,
            is_active=self.is_active,
        )



from fastapi import APIRouter, status, Header, Request
from typing import Optional, Annotated
import json

from admyral.models import WorkflowTriggerResponse, WorkflowTriggerType
from admyral.server.deps import get_admyral_store, get_workers_client
from admyral.typings import JsonValue
from admyral.utils.collections import is_not_empty


router = APIRouter()


WEBHOOK_SOURCE_NAME = "webhook"


async def _extract_payload_from_request(request: Request) -> JsonValue:
    content_type = request.headers.get("Content-Type")
    if content_type == "application/x-www-form-urlencoded":
        body = await request.form()
        # Handle Slack interactivity payload
        if (
            request.headers.get("user-agent")
            == "Slackbot 1.0 (+https://api.slack.com/robots)"
        ):
            payload = json.loads(body.get("payload"))
        else:
            raise ValueError("Unsupported content type.")
    else:
        body = await request.body()
        if is_not_empty(body):
            payload = json.loads(body)
        else:
            payload = None
    return payload


async def _handle_webhook_trigger(
    webhook_id: str,
    webhook_secret: str,
    payload: Optional[JsonValue],
) -> WorkflowTriggerResponse:
    """
    Handle the webhook trigger.

    Args:
        webhook_id: The webhook id.
    """
    payload = payload or {}
    if not isinstance(payload, dict):
        raise ValueError("Payload must be a JSON object.")

    webhook = await get_admyral_store().get_webhook(webhook_id)
    if not webhook:
        raise ValueError(f"Webhook with id {webhook_id} not found.")
    if webhook.webhook_secret != webhook_secret:
        raise ValueError("Invalid webhook secret.")
    # check whether the workflow is active
    user_id_and_workflow = await get_admyral_store().get_workflow_for_webhook(
        webhook.workflow_id
    )
    if not user_id_and_workflow:
        raise ValueError(f"Invalid webhook with id {webhook_id}. Workflow not found.")
    user_id, workflow = user_id_and_workflow
    if not workflow.is_active:
        return WorkflowTriggerResponse.inactive()

    webhook_triggers = list(
        filter(
            lambda trigger: trigger.type == WorkflowTriggerType.WEBHOOK,
            workflow.workflow_dag.start.triggers,
        )
    )
    if len(webhook_triggers) > 1:
        raise ValueError("Multiple webhook triggers found.")
    webhook_trigger = webhook_triggers[0]

    # launch the workflow execution in the background
    await get_workers_client().start_workflow(
        user_id,
        workflow,
        WEBHOOK_SOURCE_NAME,
        payload,
        trigger_default_args=webhook_trigger.default_args_dict,
    )

    return WorkflowTriggerResponse.success()


@router.post("/{webhook_id}", status_code=status.HTTP_200_OK)
async def trigger_webhook_post(
    webhook_id: str,
    request: Request,
    authorization: Annotated[str | None, Header()] = None,
) -> WorkflowTriggerResponse:
    """
    Trigger the webhook with the given id.

    Args:
        webhook_id: The webhook id.
    """
    payload = await _extract_payload_from_request(request)
    return await _handle_webhook_trigger(webhook_id, authorization, payload)


@router.post("/{webhook_id}/{webhook_secret}", status_code=status.HTTP_200_OK)
async def trigger_webhook_post_with_secret_path(
    webhook_id: str,
    webhook_secret: str,
    request: Request,
    # payload: Annotated[JsonValue, Body()] = None,
) -> WorkflowTriggerResponse:
    """
    Trigger the webhook with the given id.

    Args:
        webhook_id: The webhook id.
    """
    payload = await _extract_payload_from_request(request)
    return await _handle_webhook_trigger(webhook_id, webhook_secret, payload)


@router.get("/{webhook_id}", status_code=status.HTTP_200_OK)
async def trigger_webhook_get(
    webhook_id: str, request: Request
) -> WorkflowTriggerResponse:
    """
    Trigger the webhook with the given id.

    Args:
        webhook_id: The webhook id.
    """
    params = dict(request.query_params.items())
    authorization = request.headers.get("Authorization")
    return await _handle_webhook_trigger(webhook_id, authorization, params)


@router.get("/{webhook_id}/{webhook_secret}", status_code=status.HTTP_200_OK)
async def trigger_webhook_get_with_secret_path(
    webhook_id: str,
    webhook_secret: str,
    request: Request,
) -> WorkflowTriggerResponse:
    """
    Trigger the webhook with the given id.

    Args:
        webhook_id: The webhook id.
    """
    params = dict(request.query_params.items())
    return await _handle_webhook_trigger(webhook_id, webhook_secret, params)



"""introduce composed primary key for secrets table

Revision ID: 7e5d82221765
Revises: db5cf73c0869
Create Date: 2024-11-05 10:52:24.394416

"""

from typing import Sequence, Union

from alembic import op
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "7e5d82221765"
down_revision: Union[str, None] = "db5cf73c0869"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    op.execute("""
        ALTER TABLE secrets
        DROP CONSTRAINT secrets_pkey,
        ADD PRIMARY KEY (secret_id, user_id)
    """)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    conn = op.get_bind()

    # Check for duplicates first
    result = conn.execute(
        sqlmodel.text("""
        SELECT COUNT(*) 
        FROM (
            SELECT secret_id
            FROM secrets
            GROUP BY secret_id
            HAVING COUNT(*) > 1
        ) as duplicates
    """)
    )
    duplicate_count = result.scalar()

    if duplicate_count > 0:
        raise Exception(
            "Cannot downgrade: Multiple secrets exist with the same secret_id"
        )

    op.execute("""
        ALTER TABLE secrets
        DROP CONSTRAINT secrets_pkey,
        ADD PRIMARY KEY (secret_id)
    """)

    # ### end Alembic commands ###



"""change default user email

Revision ID: fa67e2abfd00
Revises: 2c07323c94bf
Create Date: 2024-11-03 16:58:03.800495

"""

from typing import Sequence, Union

from alembic import op
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "fa67e2abfd00"
down_revision: Union[str, None] = "2c07323c94bf"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


DEFAULT_USER_ID = "38815447-e272-4299-94c0-29a2d30435f9"


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    op.execute(
        f"UPDATE \"User\" SET email = 'default.user@admyral.ai' WHERE id = '{DEFAULT_USER_ID}'",
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    op.execute(
        f"UPDATE \"User\" SET email = 'random@email.com' WHERE id = '{DEFAULT_USER_ID}'",
    )

    # ### end Alembic commands ###



from typing import Annotated
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="1Password")
class OnePasswordSecret(BaseModel):
    domain: str
    api_key: str


def get_1password_client(secret: OnePasswordSecret) -> Client:
    return Client(
        base_url=f"https://{secret.domain}",
        headers={
            "Authorization": f"Bearer {secret.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="List Audit Events",
    display_namespace="1Password",
    description="List audit events from 1Password",
    secrets_placeholders=["1PASSWORD_SECRET"],
)
def list_1password_audit_events(
    action_type_filter: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Action Type Filter",
            description="The action type to filter events by.",
        ),
    ] = None,
    object_type_filter: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Object Type Filter",
            description="The object type to filter events by.",
        ),
    ] = None,
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The date and time to start retrieving events in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ). "
            "If not specified, start_time will default to one hour before specified end_time. If no end_time is "
            "specified, start_time will default to one hour ago.",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The date and time to stop retrieving events in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of events to retrieve.",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    # https://developer.1password.com/docs/events-api/reference#post-apiv1auditevents

    secret = ctx.get().secrets.get("1PASSWORD_SECRET")
    secret = OnePasswordSecret.model_validate(secret)

    with get_1password_client(secret) as client:
        events = []

        body = {
            "limit": 1000,
        }
        if start_time:
            body["start_time"] = (
                f"{start_time[:-1]}+00:00" if start_time.endswith("Z") else start_time
            )
        if end_time:
            body["end_time"] = (
                f"{end_time[:-1]}+00:00" if end_time.endswith("Z") else end_time
            )

        while limit is None or len(events) < limit:
            client_response = client.post("/api/v1/auditevents", json=body)
            client_response.raise_for_status()

            data = client_response.json()

            if action_type_filter and object_type_filter:
                items = [
                    event
                    for event in data["items"]
                    if event["action"] == action_type_filter
                    and event["object_type"] == object_type_filter
                ]
            elif action_type_filter:
                items = [
                    event
                    for event in data["items"]
                    if event["action"] == action_type_filter
                ]
            elif object_type_filter:
                items = [
                    event
                    for event in data["items"]
                    if event["object_type"] == object_type_filter
                ]
            else:
                items = data["items"]

            events.extend(items)

            if data["has_more"]:
                body = {"cursor": data["cursor"]}
            else:
                break

        return events if limit is None else events[:limit]



from logging.config import fileConfig

import asyncio
import os
from sqlalchemy.engine import Connection, create_engine
from sqlalchemy.ext.asyncio import async_engine_from_config, create_async_engine
from sqlalchemy import pool
from sqlmodel import SQLModel, text
from dotenv import load_dotenv

from alembic import context

from admyral.db.schemas import *  # noqa F403

load_dotenv()

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
# target_metadata = None
target_metadata = SQLModel.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    db_url = config.get_main_option(
        "sqlalchemy.url", os.environ["ADMYRAL_DATABASE_URL"]
    )
    # if url.startswith("postgresql://"):
    #     url = url.replace("postgresql://", "postgresql+asyncpg://")

    # create database
    orig_db_name = db_url.split("/")[-1]
    postgres_db_url = db_url[: -len(orig_db_name)] + "postgres"
    engine = create_engine(postgres_db_url, echo=True, future=True, pool_pre_ping=True)
    with engine.connect() as conn:
        result = conn.execute(
            text(
                f"select exists (select 1 from pg_database where datname = '{orig_db_name}')"
            )
        )
        if not result.scalar():
            conn.execute(text("commit"))
            conn.execute(text(f"create database {orig_db_name}"))

    context.configure(
        url=db_url,
        target_metadata=target_metadata,
        version_table_schema=target_metadata.schema,
        literal_binds=True,
        include_schemas=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: Connection) -> None:
    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        version_table_schema=target_metadata.schema,
        include_schemas=True,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations() -> None:
    """In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    cfg = config.get_section(config.config_ini_section, {})
    cfg["sqlalchemy.url"] = cfg.get(
        "sqlalchemy.url", os.environ.get("ADMYRAL_DATABASE_URL")
    )
    if cfg["sqlalchemy.url"].startswith("postgresql://"):
        cfg["sqlalchemy.url"] = cfg["sqlalchemy.url"].replace(
            "postgresql://", "postgresql+asyncpg://"
        )

    # create database
    db_url = cfg["sqlalchemy.url"]
    orig_db_name = db_url.split("/")[-1]
    postgres_db_url = db_url[: -len(orig_db_name)] + "postgres"
    engine = create_async_engine(
        postgres_db_url, echo=True, future=True, pool_pre_ping=True
    )
    async with engine.connect() as conn:
        result = await conn.execute(
            text(
                f"select exists (select 1 from pg_database where datname = '{orig_db_name}')"
            )
        )
        if not result.scalar():
            await conn.execute(text("commit"))
            await conn.execute(text(f"create database {orig_db_name}"))

    connectable = async_engine_from_config(
        cfg,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)

    await connectable.dispose()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode."""

    asyncio.run(run_async_migrations())


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from typing import Annotated
from openai import AzureOpenAI
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.secret.secret import register_secret


@register_secret(secret_type="Azure OpenAI")
class AzureOpenAISecret(BaseModel):
    endpoint: str
    api_key: str
    deployment_name: str


@action(
    display_name="Chat Completion",
    display_namespace="Azure OpenAI",
    description="Use advanced AI models from Azure OpenAI to perform complex tasks, such as categorization, "
    "analysis, summarization, or decision support.",
    secrets_placeholders=["AZURE_OPENAI_SECRET"],
)
def azure_openai_chat_completion(
    prompt: Annotated[
        str,
        ArgumentMetadata(
            display_name="Prompt",
            description="The prompt to use for the chat completion.",
        ),
    ],
    top_p: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Top P",
            description="Value between 0 and 1. An alternative to sampling with temperature, "
            "called nucleus sampling, where the model considers the results of the tokens "
            "with Top P probability mass. So 0.1 means only the tokens comprising the top 10% "
            "probability mass are considered. We generally recommend altering this or temperature "
            "but not both.",
        ),
    ] = None,
    temperature: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Temperature",
            description="What sampling temperature to use, between 0 and 2. Higher values like 0.8 "
            "will make the output more random, while lower values like 0.2 will make it more focused "
            "and deterministic. We generally recommend altering this or Top P but not both.",
        ),
    ] = None,
    stop_tokens: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Stop Tokens",
            description="A list of tokens at which to stop the completion.",
        ),
    ] = None,
) -> str:
    # https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/chatgpt?tabs=python-new#work-with-chat-completion-models
    # TODO: add authentication via Entra ID: https://github.com/openai/openai-python/blob/main/examples/azure_ad.py
    # TODO: error handling
    secret = ctx.get().secrets.get("AZURE_OPENAI_SECRET")
    secret = AzureOpenAISecret.model_validate(secret)

    client = AzureOpenAI(
        api_version="2024-06-01", azure_endpoint=secret.endpoint, api_key=secret.api_key
    )

    model_params = {}
    if top_p is not None:
        model_params["top_p"] = top_p
    if temperature is not None:
        model_params["temperature"] = temperature
    if stop_tokens is not None and not secret.deployment_name.startswith("o1"):
        model_params["stop"] = stop_tokens

    chat_completion = client.chat.completions.create(
        messages=[{"role": "user", "content": prompt}],
        model=secret.deployment_name,
        **model_params,
    )
    return chat_completion.choices[0].message.content



from typing import Annotated
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.exceptions import NonRetryableActionError
from admyral.secret.secret import register_secret


@register_secret(secret_type="LeakCheck")
class LeakCheckSecret(BaseModel):
    api_key: str


def _get_leakcheck_v2_client(secret: LeakCheckSecret) -> Client:
    return Client(
        base_url="https://leakcheck.io/api/v2",
        headers={
            "X-API-Key": secret.api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


def _get_leakcheck_public_client() -> Client:
    return Client(
        base_url="https://leakcheck.io/api",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Lookup API v2",
    display_namespace="LeakCheck",
    description="Perform a lookup query.",
    secrets_placeholders=["LEAKCHECK_SECRET"],
)
def leakcheck_v2_lookup(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The main value to search (email, username, etc.).",
        ),
    ],
    query_type: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Query Type",
            description="The type of query to perform (email, username, etc.).",
        ),
    ] = None,
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The number of results to return. Max: 1000.",
        ),
    ] = 100,
) -> list[dict[str, JsonValue]]:
    secret = ctx.get().secrets.get("LEAKCHECK_SECRET")
    secret = LeakCheckSecret.model_validate(secret)

    if limit > 1000:
        raise ValueError("Limit cannot be greater than 1000.")

    with _get_leakcheck_v2_client(secret) as client:
        params = {"limit": limit}
        if query_type:
            params["type"] = query_type

        response = client.get(f"/query/{query}", params=params)
        response.raise_for_status()
        data = response.json()
        if not data.get("success", False):
            raise NonRetryableActionError(
                f"API responded with an error: {data.get("error", "Unknown error")}"
            )
        return data["data"]


@action(
    display_name="Lookup public API",
    display_namespace="LeakCheck",
    description="Perform a lookup query.",
)
def leakcheck_public_lookup(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The main value to search (email, username, etc.).",
        ),
    ],
) -> list[dict[str, JsonValue]]:
    with _get_leakcheck_public_client() as client:
        response = client.get(
            f"/public?check={query}",
        )
        response.raise_for_status()
        return response.json()



from admyral.actions.integrations.cloud.aws import (
    steampipe_query_aws,
    aws_s3_bucket_logging_enabled,
    aws_s3_bucket_enforce_ssl,
    aws_s3_restrict_public_read_access,
    aws_s3_restrict_public_write_access,
    aws_s3_default_encryption_enabled,
    aws_list_iam_users,
)

__all__ = [
    "steampipe_query_aws",
    "aws_s3_bucket_logging_enabled",
    "aws_s3_bucket_enforce_ssl",
    "aws_s3_restrict_public_read_access",
    "aws_s3_restrict_public_write_access",
    "aws_s3_default_encryption_enabled",
    "aws_list_iam_users",
]



from docker import DockerClient
import shutil


def is_docker_running() -> bool:
    """
    Check for running Docker daemon.

    Returns:
        bool: True if Docker daemon is running, False otherwise.
    """
    try:
        docker_client = DockerClient.from_env()
        docker_client.ping()
        return True
    except Exception:
        return False


def get_docker_compose_cmd() -> list[str]:
    if shutil.which("docker-compose") is not None:
        return ["docker-compose"]
    return ["docker", "compose"]


def list_running_docker_containers() -> list[str]:
    docker_client = DockerClient.from_env()
    containers = docker_client.containers.list()
    return [container.name for container in containers]


def clean_up_old_images(
    repository_name: str, current_version: str | None = None
) -> None:
    docker_client = DockerClient.from_env()
    images = docker_client.images.list(name=repository_name)
    for image in images:
        if current_version and f"{repository_name}:{current_version}" in image.tags:
            continue
        image.remove(force=True)



from admyral.actions.integrations.iam.okta import (
    list_okta_events,
    okta_search_users,
    okta_get_all_user_types,
    get_okta_logs,
)

__all__ = [
    "list_okta_events",
    "okta_search_users",
    "okta_get_all_user_types",
    "get_okta_logs",
]



from fastapi import APIRouter, status, Depends

from admyral.server.auth import authenticate
from admyral.models import (
    Secret,
    SecretMetadata,
    DeleteSecretRequest,
    AuthenticatedUser,
)
from admyral.server.deps import get_secrets_manager
from admyral.secret.secret_registry import SecretRegistry


def handle_secret_values(secret_key_values: dict[str, str]) -> dict[str, str]:
    return dict(
        [
            (key, val.encode("utf-8").decode("unicode_escape"))
            for key, val in secret_key_values.items()
        ]
    )


router = APIRouter()


@router.post("/set", status_code=status.HTTP_201_CREATED)
async def set_secret(
    secret: Secret, authenticated_user: AuthenticatedUser = Depends(authenticate)
) -> SecretMetadata:
    secret.secret = handle_secret_values(secret.secret)
    return await get_secrets_manager().set(authenticated_user.user_id, secret)


@router.post("/update", status_code=status.HTTP_200_OK)
async def update_secret(
    secret: Secret, authenticated_user: AuthenticatedUser = Depends(authenticate)
) -> SecretMetadata:
    secret.secret = handle_secret_values(secret.secret)
    return await get_secrets_manager().update(authenticated_user.user_id, secret)


@router.get("/list", status_code=status.HTTP_200_OK)
async def list_secrets(
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> list[SecretMetadata]:
    return await get_secrets_manager().list(user_id=authenticated_user.user_id)


@router.delete("/delete", status_code=status.HTTP_204_NO_CONTENT)
async def delete_secret(
    request: DeleteSecretRequest,
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> None:
    await get_secrets_manager().delete(
        user_id=authenticated_user.user_id, secret_id=request.secret_id
    )


@router.get("/schemas")
async def get_secret_schemas(
    _authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> list[tuple[str, list[str]]]:
    schemas = SecretRegistry.get_secret_schemas()
    schemas = list(schemas.items())
    return sorted(schemas, key=lambda x: x[0])



import click
import os
from enum import Enum
import logging
from dotenv import load_dotenv
from pydantic import BaseModel
import uuid
import yaml

load_dotenv()


class LoggingLevel(Enum):
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


ENV_ADMYRAL_APP_DIR = "ADMYRAL_APP_DIR"
ENV_ADMYRAL_LOGGING_LEVEL = "ADMYRAL_LOGGING_LEVEL"
ENV_ADMYRAL_DISABLE_NSJAIL = "ADMYRAL_DISABLE_NSJAIL"
ENV_ADMYRAL_FLOCK_PATH = "ADMYRAL_FLOCK_PATH"
ENV_ADMYRAL_PIP_CACHE_DIRECTORY = "ADMYRAL_PIP_CACHE_DIRECTORY"
ENV_ADMYRAL_PIP_LOCK_CACHE_DIRECTORY = "ADMYRAL_PIP_LOCK_CACHE_DIRECTORY"
ENV_ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE = "ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE"
ENV_ADMYRAL_ENV = "ADMYRAL_ENV"
ENV_ADMYRAL_POSTHOG_API_KEY = "ADMYRAL_POSTHOG_API_KEY"
ENV_ADMYRAL_POSTHOG_HOST = "ADMYRAL_POSTHOG_HOST"
ENV_ADMYRAL_DISABLE_TELEMETRY = "ADMYRAL_DISABLE_TELEMETRY"

APP_NAME = "Admyral"

# Name of the configuration file
GLOBAL_CONFIG_FILE = "config.yaml"

PROJECT_DIRECTORY = ".admyral"

ADMYRAL_DEFAULT_STORAGE_DIRECTORY_NAME = "admyral_default_storage"

LOGGING_LEVEL = os.getenv(ENV_ADMYRAL_LOGGING_LEVEL, "INFO").upper()


def get_global_project_directory() -> str:
    """
    # TODO: adapt descr.
    Returns the global config folder for Admyral. The default behavior is to return whatever is most appropriate for the operating system.
    """
    return os.environ.get(ENV_ADMYRAL_APP_DIR, click.get_app_dir(APP_NAME))


def get_local_storage_path() -> str:
    """
    TODO: add descr.
    """
    return os.path.join(
        get_global_project_directory(), ADMYRAL_DEFAULT_STORAGE_DIRECTORY_NAME
    )


def get_local_postgres_volume() -> str:
    return os.path.join(get_local_storage_path(), "postgres")


def get_user_config_file() -> str:
    return os.path.join(get_global_project_directory(), GLOBAL_CONFIG_FILE)


# Constants for custom Python execution
ADMYRAL_DISABLE_NSJAIL = (
    os.getenv(ENV_ADMYRAL_DISABLE_NSJAIL, "false").lower() == "true"
)
ADMYRAL_CACHE_DIRECOTRY = os.path.join(get_local_storage_path(), "cache")
ADMYRAL_PIP_CACHE_DIRECTORY = os.getenv(
    ENV_ADMYRAL_PIP_CACHE_DIRECTORY,
    os.path.join(ADMYRAL_CACHE_DIRECOTRY, "pip"),
)
ADMYRAL_PIP_LOCK_CACHE_DIRECTORY = os.getenv(
    ENV_ADMYRAL_PIP_LOCK_CACHE_DIRECTORY,
    os.path.join(ADMYRAL_CACHE_DIRECOTRY, "pip-lock"),
)
ADMYRAL_PIP_LOCKFILE_CACHE_TTL_IN_SECONDS = 3 * 24 * 60 * 60  # 3 days
ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE = (
    os.getenv(ENV_ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE, "true").lower() == "true"
)
ADMYRAL_ENV = os.getenv(ENV_ADMYRAL_ENV, "prod")
ADMYRAL_POSTHOG_API_KEY = os.getenv(
    ENV_ADMYRAL_POSTHOG_API_KEY, "phc_RIpkRea4KLW6EONEDCSZVR1Td4YzeHf4ziUsGzmPnjD"
)
ADMYRAL_POSTHOG_HOST = os.getenv(ENV_ADMYRAL_POSTHOG_HOST, "https://eu.i.posthog.com")

ADMYRAL_DISABLE_TELEMETRY = (
    os.getenv(ENV_ADMYRAL_DISABLE_TELEMETRY, "false").lower() == "true"
)

# DATABASE AND SECRETS MANAGER


class DatabaseType(str, Enum):
    """
    Enum class for the supported database types.
    """

    POSTGRES = "postgres"


class SecretsManagerType(str, Enum):
    """
    Enum class for the supported secrets manager types.
    """

    SQL = "sql"
    # AWS_SECRETS_MANAGER = "aws_secrets_manager"


ENV_ADMYRAL_DATABASE_URL = "ADMYRAL_DATABASE_URL"
ENV_ADMYRAL_SECRETS_MANAGER_TYPE = "ADMYRAL_SECRETS_MANAGER"


ADMYRAL_DATABASE_URL = os.getenv(
    ENV_ADMYRAL_DATABASE_URL,
    "postgresql://postgres:your-super-secret-and-long-postgres-password@localhost:5432/admyral",
)
if ADMYRAL_DATABASE_URL.startswith("postgresql"):
    if ADMYRAL_DATABASE_URL.startswith("postgresql://"):
        ADMYRAL_DATABASE_URL = ADMYRAL_DATABASE_URL.replace(
            "postgresql://", "postgresql+asyncpg://"
        )
    ADMYRAL_DATABASE_TYPE = DatabaseType.POSTGRES
else:
    raise NotImplementedError(f"Unsupported database type: {ADMYRAL_DATABASE_URL}")

ADMYRAL_SECRETS_MANAGER_TYPE = SecretsManagerType(
    os.getenv(ENV_ADMYRAL_SECRETS_MANAGER_TYPE, SecretsManagerType.SQL)
)


ENV_TEMPORAL_HOST = "ADMYRAL_TEMPORAL_HOST"

ADMYRAL_TEMPORAL_HOST = os.getenv(ENV_TEMPORAL_HOST, "localhost:7233")


class GlobalConfig(BaseModel):
    """
    The global configuration for Admyral.
    """

    id: str
    default_user_id: str = "38815447-e272-4299-94c0-29a2d30435f9"
    default_user_email: str = "default.user@admyral.ai"
    telemetry_disabled: bool = ADMYRAL_DISABLE_TELEMETRY
    storage_directory: str = get_local_storage_path()
    database_type: DatabaseType = ADMYRAL_DATABASE_TYPE
    database_url: str = ADMYRAL_DATABASE_URL
    temporal_host: str = ADMYRAL_TEMPORAL_HOST
    secrets_manager_type: SecretsManagerType = ADMYRAL_SECRETS_MANAGER_TYPE
    posthog_api_key: str = ADMYRAL_POSTHOG_API_KEY
    posthog_host: str = ADMYRAL_POSTHOG_HOST
    environment: str = ADMYRAL_ENV

    cli_target: str = "http://localhost:8000"
    api_key: str | None = None

    pip_lockfile_cache_cleanup_interval: int = 60 * 60 * 24  # 1 day


def load_local_config() -> GlobalConfig:
    """
    Load the global configuration for Admyral.
    """
    config_file = get_user_config_file()

    if os.path.exists(config_file):
        with open(config_file, "r") as f:
            file_content = yaml.safe_load(f)
        if "id" not in file_content:
            file_content["id"] = str(uuid.uuid4())
            with open(config_file, "w") as f:
                # Reset the file content and overwrite with new config
                yaml.dump(file_content, f)
    else:
        os.makedirs(get_local_storage_path(), exist_ok=True)
        id = str(uuid.uuid4())
        file_content = {"id": id}
        with open(config_file, "w") as f:
            yaml.dump(file_content, f)

    return GlobalConfig(**file_content)


CONFIG = load_local_config()

ADMYRAL_DAEMON_PID_FILE = "admyral.pid"
ADMYRAL_DAEMON_LOG_FILE = "admyral.log"

TEMPORAL_DAEMON_PID_FILE = "temporal.pid"
TEMPORAL_DAEMON_LOG_FILE = "temporal.log"


def get_admyral_daemon_directory() -> str:
    return os.path.join(get_global_project_directory(), "Admyral", "daemon")


def get_admyral_daemon_pid_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), ADMYRAL_DAEMON_PID_FILE)


def get_temporal_daemon_pid_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), TEMPORAL_DAEMON_PID_FILE)


def get_admyral_daemon_log_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), ADMYRAL_DAEMON_LOG_FILE)


def get_temporal_daemon_log_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), TEMPORAL_DAEMON_LOG_FILE)


ENV_ADMYRAL_WEBHOOK_SIGNING_SECRET = "ADMYRAL_WEBHOOK_SIGNING_SECRET"
WEBHOOK_SIGNING_SECRET = bytes.fromhex(
    os.getenv(
        ENV_ADMYRAL_WEBHOOK_SIGNING_SECRET,
        "ebedd9f0c10b01acb9fd097cdf34a2b38bb554f2c7f68f0d9f534eff5c6ef5d9",
    )
)

ENV_ADMYRAL_SECRETS_ENCRYPTION_KEY = "ADMYRAL_SECRETS_ENCRYPTION_KEY"
SECRETS_ENCRYPTION_KEY = bytes.fromhex(
    os.getenv(
        ENV_ADMYRAL_SECRETS_ENCRYPTION_KEY,
        "834f01cf391c972f1e6def3d7f315f8194bb10048e5cf282aa4cba63b239d8fb",
    )
)


API_V1_STR = "/api/v1"


ENV_ADMYRAL_DISABLE_AUTH = "ADMYRAL_DISABLE_AUTH"
DISABLE_AUTH = os.getenv(ENV_ADMYRAL_DISABLE_AUTH, "false").lower() == "true"


ENV_ADMYRAL_AUTH_SECRET = "NEXTAUTH_SECRET"
AUTH_SECRET = os.getenv(
    ENV_ADMYRAL_AUTH_SECRET, "QzkuVCn7OGfkpoX98aOxf2tc3kFX8pZs71N1wHPo8NM="
)


TEST_USER_ID = "a2f038f1-e35b-4509-bcc4-c08bd0e481a6"



from abc import ABC, abstractmethod

from admyral.models import (
    User,
    PipLockfile,
    PythonAction,
    ActionMetadata,
    Workflow,
    WorkflowRun,
    WorkflowWebhook,
    WorkflowSchedule,
    EncryptedSecret,
    WorkflowMetadata,
    SecretMetadata,
    WorkflowRunMetadata,
    WorkflowRunStepMetadata,
    WorkflowRunStep,
    ApiKey,
)
from admyral.typings import JsonValue


class StoreInterface(ABC):
    ########################################################
    # User Management
    ########################################################

    @abstractmethod
    async def get_user(self, user_id: str) -> User | None: ...

    @abstractmethod
    async def clean_up_workflow_data_of(self, user_id: str) -> None: ...

    @abstractmethod
    async def delete_user(self, user_id: str) -> None: ...

    ########################################################
    # API Key Management
    ########################################################

    @abstractmethod
    async def store_api_key(self, user_id: str, name: str, key: str) -> ApiKey: ...

    @abstractmethod
    async def list_api_keys(self, user_id: str) -> list[ApiKey]: ...

    @abstractmethod
    async def search_api_key_owner(self, key: str) -> str | None: ...

    @abstractmethod
    async def delete_api_key(self, user_id: str, key_id: str) -> None: ...

    ########################################################
    # Python Action
    ########################################################

    @abstractmethod
    async def list_actions(self, user_id: str) -> list[ActionMetadata]: ...

    @abstractmethod
    async def get_action(
        self, user_id: str, action_type: str
    ) -> PythonAction | None: ...

    @abstractmethod
    async def store_action(self, user_id: str, action: PythonAction) -> None: ...

    @abstractmethod
    async def delete_action(self, user_id: str, action_type: str) -> None: ...

    ########################################################
    # Pip Lockfile Cache
    ########################################################

    @abstractmethod
    async def get_cached_pip_lockfile(self, hash: str) -> PipLockfile | None: ...

    @abstractmethod
    async def cache_pip_lockfile(self, pip_lockfile: PipLockfile) -> None: ...

    @abstractmethod
    async def delete_expired_cached_pip_lockfile(self) -> None: ...

    ########################################################
    # Workflows
    ########################################################

    @abstractmethod
    async def list_workflows(self, user_id: str) -> list[WorkflowMetadata]: ...

    @abstractmethod
    async def get_workflow_by_name(
        self, user_id: str, workflow_name: str
    ) -> Workflow | None: ...

    @abstractmethod
    async def get_workflow_by_id(
        self, user_id: str, workflow_id: str
    ) -> Workflow | None: ...

    @abstractmethod
    async def store_workflow(self, user_id: str, workflow: Workflow) -> None: ...

    @abstractmethod
    async def set_workflow_active_state(
        self, user_id: str, workflow_id: str, is_active: bool
    ) -> None: ...

    @abstractmethod
    async def remove_workflow(self, user_id: str, workflow_id: str) -> None: ...

    @abstractmethod
    async def get_workflow_for_webhook(
        self, workflow_id: str
    ) -> tuple[str, Workflow] | None: ...

    ########################################################
    # Workflow Webhooks
    ########################################################

    @abstractmethod
    async def store_workflow_webhook(
        self, user_id: str, workflow_id: str
    ) -> WorkflowWebhook: ...

    @abstractmethod
    async def get_webhook_for_workflow(
        self, user_id: str, workflow_id: str
    ) -> WorkflowWebhook | None: ...

    @abstractmethod
    async def get_webhook(self, webhook_id: str) -> WorkflowWebhook | None: ...

    @abstractmethod
    async def delete_webhook(self, user_id: str, webhook_id: str) -> None: ...

    ########################################################
    # Workflow Schedules
    ########################################################

    @abstractmethod
    async def store_schedule(
        self, user_id: str, schedule: WorkflowSchedule
    ) -> None: ...

    @abstractmethod
    async def list_schedules_for_workflow(
        self, user_id: str, workflow_id: str
    ) -> list[WorkflowSchedule]: ...

    @abstractmethod
    async def delete_schedule(self, user_id: str, schedule_id: str) -> None: ...

    ########################################################
    # Workflow Runs - User Facing
    ########################################################

    @abstractmethod
    async def list_workflow_runs(
        self, user_id: str, workflow_id: str, limit: int = 100
    ) -> list[WorkflowRunMetadata]: ...

    @abstractmethod
    async def get_workflow_run(
        self, workflow_id: str, run_id: str
    ) -> WorkflowRun | None: ...

    @abstractmethod
    async def list_workflow_run_steps(
        self, user_id: str, workflow_id: str, run_id: str
    ) -> list[WorkflowRunStepMetadata]: ...

    @abstractmethod
    async def get_workflow_run_step(
        self, user_id: str, workflow_id: str, run_id: str, step_id: str
    ) -> WorkflowRunStep | None: ...

    ########################################################
    # Workflow Runs - State Updates during execution
    ########################################################

    @abstractmethod
    async def init_workflow_run(
        self,
        run_id: str,
        step_id: str,
        workflow_id: str,
        source_name: str,
        payload: dict[str, JsonValue],
    ) -> None: ...

    @abstractmethod
    async def mark_workflow_run_as_completed(
        self, run_id: str, completed_at: str
    ) -> None: ...

    @abstractmethod
    async def append_logs(
        self, step_id: str, run_id: str, action_id: str, prev_step_id: str, logs: str
    ) -> None: ...

    @abstractmethod
    async def store_action_result(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        result: JsonValue,
        input_args: dict[str, JsonValue],
    ) -> None: ...

    @abstractmethod
    async def store_workflow_run_error(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        error: str,
        input_args: dict[str, JsonValue],
    ) -> None: ...

    ########################################################
    # Secrets
    ########################################################

    @abstractmethod
    async def list_secrets(
        self, user_id: str, secret_type: str | None = None
    ) -> list[SecretMetadata]: ...

    @abstractmethod
    async def get_secret(
        self, user_id: str, secret_id: str
    ) -> EncryptedSecret | None: ...

    @abstractmethod
    async def store_secret(
        self,
        user_id: str,
        secret_id: str,
        encrypted_secret: str | None,
        secret_schema: list[str],
        secret_type: str | None = None,
    ) -> SecretMetadata: ...

    @abstractmethod
    async def delete_secret(self, user_id: str, secret_id: str) -> None: ...

    @abstractmethod
    async def compare_and_swap_secret(
        self,
        user_id: str,
        secret_id: str,
        expected_encrypted_secret: str | None,
        encrypted_secret: str | None,
        secret_schema: list[str],
        secret_type: str | None = None,
    ) -> SecretMetadata: ...
